#import os
import datetime
#import numpy as np
#from open_ephys.control import OpenEphysHTTPServer
import requests
import json
from riglib.open_ephys import OpenEphysHTTPServer

IP_neuropixel = '10.155.205.108'
address = f'http://{IP_neuropixel}:37497'

gui = OpenEphysHTTPServer(IP_neuropixel)
parent_dir = 'E://Neuropixel_data'

prepend_text = str(datetime.date.today())
filename = '_Neuropixel_'
append_text = 'test6'

gui.set_prepend_text(prepend_text)
gui.set_base_text(filename)
gui.set_append_text(append_text)
gui.acquire()

#gui.set_start_new_dir() # This isn't necessary because saved directory changes when file name changes
#gui.record(10)

# gui.set_parent_dir(parent_dir)
# print(gui.get_recording_info())
#print(gui.get_processors('Neuropix-PXI'))
#A = gui.get_parameters(105,0)
#print(A)
#print(gui.get_parameters(103,1))


"""
Robot manipulability operations.

@author: Peter Corke
@copyright: Peter Corke
"""



from numpy import *
from .jacobian import jacob0
from .dynamics import inertia
from numpy.linalg import inv,eig,det,svd
from .utility import *

def manipulability(robot, q, which = 'yoshikawa'):
    '''
    MANIPLTY Manipulability measure

        - M = MANIPLTY(ROBOT, Q)
        - M = MANIPLTY(ROBOT, Q, WHICH)

    Computes the manipulability index for the manipulator at the given pose.

    For an n-axis manipulator Q may be an n-element vector, or an m x n
    joint space trajectory.

    If Q is a vector MANIPLTY returns a scalar manipulability index.
    If Q is a matrix MANIPLTY returns a column vector of  manipulability 
    indices for each pose specified by Q.

    The argument WHICH can be either 'yoshikawa' (default) or 'asada' and
    selects one of two manipulability measures.
    Yoshikawa's manipulability measure gives an indication of how far 
    the manipulator is from singularities and thus able to move and 
    exert forces uniformly in all directions.

    Asada's manipulability measure is based on the manipulator's
    Cartesian inertia matrix.  An n-dimensional inertia ellipsoid
        X' M(q) X = 1
    gives an indication of how well the manipulator can accelerate
    in each of the Cartesian directions.  The scalar measure computed
    here is the ratio of the smallest/largest ellipsoid axis.  Ideally
    the ellipsoid would be spherical, giving a ratio of 1, but in
    practice will be less than 1.

    @see: inertia, jacob0
    '''        
    n = robot.n
    q = mat(q)
    w = array([])
    if 'yoshikawa'.startswith(which):
            if numrows(q)==1:
                    return yoshi(robot,q)
            for Q in q:
                    w = concatenate((w,array([yoshi(robot,Q)])))
    elif 'asada'.startswith(which):
            if numrows(q)==1:
                    return asada(robot,q)
            for Q in q:
                    w = concatenate((w,array([asada(robot,Q)])))
    return mat(w)

def yoshi(robot,q):
        J = jacob0(robot,q)
        return sqrt(det(J*J.T))

def asada(robot,q):
        J = jacob0(robot,q)
        Ji = inv(J)
        M = inertia(robot,q)
        Mx = Ji.T*M*Ji
        e = eig(Mx)[0]

        return real( e.min(0)/e.max(0) )



'''
A generic target tracking task
'''
import numpy as np

from riglib.experiment import traits, Sequence
from riglib import plants

from riglib.stereo_opengl.window import Window
from .target_graphics import *

## Plants
# List of possible "plants" that a subject could control either during manual or brain control
cursor = plants.CursorPlant()
shoulder_anchor = np.array([2., 0., -15])
chain_kwargs = dict(link_radii=.6, joint_radii=0.6, joint_colors=(181/256., 116/256., 96/256., 1), link_colors=(181/256., 116/256., 96/256., 1))
chain_20_20_endpt = plants.EndptControlled2LArm(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)
chain_20_20 = plants.RobotArmGen2D(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)

plantlist = dict(
    cursor=cursor,
    chain_20_20=chain_20_20,
    chain_20_20_endpt=chain_20_20_endpt)

class TargetTracking(Sequence):
    '''
    This is a generic cued target tracking skeleton, to form as a common ancestor to the most
    common type of motor control tracking tasks.
    '''
    status = dict(
        wait = dict(start_trial="trajectory", start_pause="pause"),
        wait_retry = dict(start_trial="trajectory", start_pause="pause"),
        trajectory = dict(enter_target="hold", timeout="timeout_penalty", start_pause="pause"),
        hold = dict(hold_complete_go_ramp="tracking_in_ramp", hold_complete_no_ramp="tracking_in", leave_target="hold_penalty", start_pause="pause"),

        tracking_in_ramp = dict(ramp_complete="tracking_in", ramp_and_trial_complete="reward", leave_target="tracking_out_ramp", start_pause="pause"),
        tracking_out_ramp = dict(ramp_complete="tracking_out", ramp_and_trial_complete="reward", enter_target="tracking_in_ramp", tracking_out_timeout="tracking_out_penalty", start_pause="pause"),
        
        tracking_in = dict(traj_complete="tracking_in_ramp", trial_complete="reward", leave_target="tracking_out", start_pause="pause"),
        tracking_out = dict(traj_complete="tracking_out_ramp", trial_complete="reward", enter_target="tracking_in", tracking_out_timeout="tracking_out_penalty", start_pause="pause"),
        
        timeout_penalty = dict(timeout_penalty_end="wait", start_pause="pause", end_state=True),
        hold_penalty = dict(hold_penalty_end="wait", hold_penalty_end_retry="wait_retry", start_pause="pause", end_state=True),
        tracking_out_penalty = dict(tracking_out_penalty_end="wait", start_pause="pause", end_state=True),
        reward = dict(reward_end="wait", start_pause="pause", stoppable=False, end_state=True),
        pause = dict(end_pause="wait", end_state=True)
        # all end_states will result in trial counter +1, so if you start pause during a penalty state, 
        # the next trial after unpausing will be current trial +2

    )

    # initial state
    state = "wait"
    tries = 0 # Helper variable to keep track of the number of failed attempts at initiating a given trial

    wait_time = traits.Float(1., desc="Length of time in wait state (inter-trial interval)")
    reward_time = traits.Float(.5, desc="Length of reward dispensation")
    timeout_time = traits.Float(10, desc="Time allowed to go between trajectories")
    timeout_penalty_time = traits.Float(1, desc="Length of penalty time for initiation timeout error")
    hold_time = traits.Float(.1, desc="Time of hold required at target before trajectory begins")
    hold_penalty_time = traits.Float(1, desc="Length of penalty time for target hold error")
    tracking_out_time = traits.Float(2.5, desc="Time allowed to be tracking outside the target") # AKA tolerance time
    tracking_out_penalty_time = traits.Float(1, desc="Length of penalty time for tracking out error")
    max_hold_attempts = traits.Int(5, desc='Number of attempts to initiate a trial before skipping to the next one')
    tracking_reward_time = traits.Float(.15, desc="Length of reward for tracking in")
    tracking_reward_interval = traits.Float(.5, desc="Length of time between reward for tracking in (must be greater than reward time)")
    
    def init(self):
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4'), ('target', 'f8',(3,)), ('disturbance', 'f8',(3,)), ('is_disturbance', '?')])
        super().init()

        self.frame_index = 0 # index in the frame in a trial
        self.total_distance_error = 0 # Euclidian distance between cursor and target during each trial
        self.trial_timed_out = True # check if the trial is finished
        self.plant_position = []
        self.disturbance_trial = False
        self.repeat_freq_set = False
        self.gen_index = -1

        if self.velocity_control:
            print('VELOCITY CONTROL')
        else:
            print('POSITION CONTROL') # default is position control - see manualcontrolmixin
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _parse_next_trial(self):
        '''Get the required data from the generator'''
        # yield idx, targs, disturbance, dis_trajectory, sample_rate, ramp, ramp_down :
        self.gen_index, self.targs, self.disturbance_trial, self.disturbance_path, self.sample_rate, self.ramp_up_time, self.ramp_down_time = self.next_trial # targs and disturbance are same length

        self.targs = np.squeeze(self.targs,axis=0)
        self.disturbance_path = np.squeeze(self.disturbance_path)

        WIDTH, HEIGHT = self.window_size[0], self.window_size[1]
        lookahead = np.zeros((self.lookahead,np.shape(self.targs)[1])) # (30,3)

        self.targs = self.trajectory_amplitude*self.targs
        self.disturbance_path = self.disturbance_amplitude*self.disturbance_path
        # print(np.amax(self.targs), np.amax(self.disturbance_path))

        self.ramp_counter = np.zeros((len(self.targs),), dtype='int')
        if self.ramp_up_time > 0:
            self.ramp_counter[:int(self.ramp_up_time*self.sample_rate)] = 1
        if self.ramp_down_time > 0:
            self.ramp_counter[-int(self.ramp_down_time*self.sample_rate):] = 2

        self.targs = np.concatenate((lookahead, self.targs),axis=0) # (time_length*sample_rate+30,3) # targs and disturbance are no longer same length
    
    def tracking_task_start_wait(self):
        print(self.gen_index)

        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.gen_index
        self.trial_record['is_disturbance'] = self.disturbance_trial
        for i in range(len(self.disturbance_path)):
            # Update the data sinks with trial information --> bmi3d_trials
            self.trial_record['target'] = self.targs[i+self.lookahead]
            self.trial_record['disturbance'] = self.disturbance_path[i]
            self.sinks.send("trials", self.trial_record)

        # trial is not finished
        self.trial_timed_out = False

        # number of times this trajectory has been attempted
        self.tries = 0

        # index into trajectory
        self.frame_index = -1

        # number of frames in trajectory
        '''Nothing generic to do.'''
        self.trajectory_length = len(self.targs)

        # saved plant poitions
        self.plant_position = []

    def _start_wait(self):
        # Call parent method to draw the next target capture sequence from the generator
        super()._start_wait()
        if self.sample_rate != self.fps:
            print('WARNING: generator sample rate should equal FSM fps!')

        if self.repeat_freq_set:
            try:
                self.next_trial = next(self.gen)
                self._parse_next_trial()
            except StopIteration:
                self.end_task()

        self.tracking_task_start_wait()

    def _while_wait(self):
        '''Nothing generic to do.'''
        pass

    def _start_wait_retry(self):
        self.tracking_task_start_wait()

    def _while_wait_retry(self):
        '''Nothing generic to do.'''
        pass

    def _start_trajectory(self):
        self.tries += 1
        self.frame_index += 1

    def _while_trajectory(self):
        '''Nothing generic to do.'''
        pass        

    def _end_trajectory(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold(self):
        '''Nothing generic to do.'''
        pass

    def _while_hold(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_hold(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_in_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_in_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_in_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_out_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_out_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_out_ramp(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_in(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _end_tracking_out(self):
        '''Nothing generic to do.'''
        pass

    def _start_timeout_penalty(self):
        '''Nothing generic to do.'''
        # self._increment_tries()
        pass

    def _while_timeout_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_timeout_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold_penalty(self):
        '''Nothing generic to do.'''
        # self._increment_tries()
        pass

    def _while_hold_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_hold_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_tracking_out_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _while_tracking_out_penalty(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_tracking_out_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_reward(self):
        '''Nothing generic to do.'''
        pass

    def _while_reward(self):
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_reward(self):
        '''Nothing generic to do.'''
        pass

    def _start_pause(self):
        '''Nothing generic to do.'''
        pass

    def _while_pause(self):
        '''Nothing generic to do.'''
        self.pos_offset = [0,0,0]
        self.vel_offset = [0,0,0]

    def _end_pause(self):
        '''Nothing generic to do.'''
        pass

    ################## State transition test functions ##################
    def _test_start_trial(self, time_in_state):
        '''Start next trial automatically. You may want this to instead be
            - a random delay
            - require some initiation action
        '''
        return time_in_state > self.wait_time

    def _test_timeout(self, time_in_state):
        return time_in_state > self.timeout_time

    def _test_hold_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return time_in_state > self.hold_time

    def _test_hold_complete_go_ramp(self, time_in_state):
        '''Test whether the target is held long enough and whether there is a ramp up before the trajectory'''
        return (time_in_state > self.hold_time) and (self.ramp_up_time > 0)

    def _test_hold_complete_no_ramp(self, time_in_state):
        '''Test whether the target is held long enough and whether to go straight into the trajectory'''
        return (time_in_state > self.hold_time) and (self.ramp_up_time == 0)

    def _test_ramp_complete(self, time_in_state):
        '''Test whether the ramp up is finished'''
        return self.frame_index == self.ramp_up_time*self.sample_rate

    def _test_traj_complete(self, time_in_state):
        '''Test whether the trajectory is finished and whether there is a ramp down before the trial ends'''
        return (self.frame_index + self.lookahead == self.trajectory_length - self.ramp_down_time*self.sample_rate) and (self.ramp_down_time > 0)

    def _test_ramp_and_trial_complete(self, time_in_state):
        '''Test whether the ramp down is finished, ending the trial'''
        return (self.frame_index + self.lookahead == self.trajectory_length) and (self.ramp_down_time > 0)
    
    def _test_trial_complete(self, time_in_state):
        '''Test whether the trajectory is finished, ending the trial'''
        return (self.frame_index + self.lookahead == self.trajectory_length) and (self.ramp_down_time == 0)

    def _test_tracking_out_timeout(self, time_in_state):
        return time_in_state > self.tracking_out_time

    def _test_timeout_penalty_end(self, time_in_state):
        return time_in_state > self.timeout_penalty_time

    def _test_hold_penalty_end(self, time_in_state):
        return (time_in_state > self.hold_penalty_time) and (self.tries==self.max_hold_attempts)

    def _test_hold_penalty_end_retry(self, time_in_state):
        return (time_in_state > self.hold_penalty_time) and (self.tries<self.max_hold_attempts)

    def _test_tracking_out_penalty_end(self, time_in_state):
        return time_in_state > self.tracking_out_penalty_time

    def _test_reward_end(self, time_in_state):
        return time_in_state > self.reward_time

    def _test_enter_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return False

    def _test_leave_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return False

    def _test_start_pause(self, time_in_state):
        return self.pause

    def _test_end_pause(self, time_in_state):
        return not self.pause
    
    def update_report_stats(self):
        '''
        see experiment.Experiment.update_report_stats for docs
        '''
        super().update_report_stats()
        self.reportstats['Trial #'] = self.calc_trial_num()
        self.reportstats['Reward/min'] = np.round(self.calc_events_per_min('reward', 120.), decimals=2)
        

class ScreenTargetTracking(TargetTracking, Window):
    """Concrete implementation of Target Tracking task where the target is moving and
    are tracked by holding the cursor within the moving target"""

    limit2d = traits.Bool(True, desc="Limit cursor movement to 2D")
    limit1d = traits.Bool(True, desc="Limit cursor movement to 1D")

    sequence_generators = [
        'tracking_target_chain', 'tracking_target_debug', 'tracking_target_training'
    ]

    hidden_traits = ['cursor_color', 'trajectory_color', 'cursor_bounds', 'cursor_radius', 'plant_hide_rate', 'starting_pos']
    targets = []

    is_bmi_seed = True

    # Runtime settable traits
    target_radius = traits.Float(2, desc="Radius of targets in cm") #2,0.75
    trajectory_radius = traits.Float(.5, desc="Radius of targets in cm")
    trajectory_color = traits.OptionsList("gold", *target_colors, desc="Color of the trajectory", bmi3d_input_options=list(target_colors.keys()))
    target_color = traits.OptionsList("yellow", *target_colors, desc="Color of the target", bmi3d_input_options=list(target_colors.keys()))
    plant_hide_rate = traits.Float(0.0, desc='If the plant is visible, specifies a percentage of trials where it will be hidden')
    plant_type = traits.OptionsList(*plantlist, bmi3d_input_options=list(plantlist.keys()))
    plant_visible = traits.Bool(True, desc='Specifies whether entire plant is displayed or just endpoint')
    cursor_radius = traits.Float(.5, desc='Radius of cursor in cm')
    cursor_color = traits.OptionsList("pink", *target_colors, desc='Color of cursor endpoint', bmi3d_input_options=list(target_colors.keys()))
    cursor_bounds = traits.Tuple((-10., 10., 0., 0., -10., 10.), desc='(x min, x max, z min, z max, y min, y max)')
    starting_pos = traits.Tuple((5., 0., 5.), desc='Where to initialize the cursor')
    fps = traits.Float(60, desc="Rate at which the FSM is called in Hz") # originally set by class Experiment
    trajectory_amplitude = traits.Float(1, desc='Scale factor applied to the trajectory')
    disturbance_amplitude = traits.Float(1, desc='Scale factors applied to the disturbance')
    always_1d = traits.Bool(False, desc='always constrain movement to 1d')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize the plant
        if not hasattr(self, 'plant'):
            self.plant = plantlist[self.plant_type]
        self.plant.set_bounds(np.array(self.cursor_bounds))
        self.plant.set_color(target_colors[self.cursor_color])
        self.plant.set_cursor_radius(self.cursor_radius)
        self.plant_vis_prev = True
        self.cursor_vis_prev = True
        self.lookahead = 30 # number of frames to create a "lookahead" window of 0.5 seconds (half the screen)
        self.original_limit1d = self.limit1d # keep track of original settable trait
        
        if not self.always_1d:
            self.limit1d = False # allow 2d movement before center-hold initiation
        
        # Add graphics models for the plant and targets to the window
        if hasattr(self.plant, 'graphics_models'):
            for model in self.plant.graphics_models:
                self.add_model(model)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:
            # This is the center target being followed by the user
            self.target = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])

            # This is the trajectory that spans the screen
            self.trajectory = VirtualCableTarget(target_radius=self.trajectory_radius, target_color=target_colors[self.trajectory_color])

            # This is the progress bar
            self.bar = VirtualRectangularTarget(target_width=1, target_height=0, target_color=(0., 1., 0., 0.75), starting_pos=[0,0,9])
            # print('INIT TRAJ')

        # Declare any plant attributes which must be saved to the HDF file at the _cycle rate
        for attr in self.plant.hdf_attrs:
            self.add_dtype(*attr)

    def init(self):
        self.add_dtype('trial', 'u4', (1,))
        self.add_dtype('gen_idx', 'int', (1,)) # dtype needs to be able to represent -1
        self.add_dtype('plant_visible', '?', (1,))
        self.add_dtype('current_target', 'f8', (3,))
        self.add_dtype('current_disturbance', 'f8', (3,)) # see task_data['manual_input'] for cursor position without added disturbance
        self.add_dtype('current_target_validate', 'f8', (3,))
        super().init()
        self.plant.set_endpoint_pos(np.array(self.starting_pos))

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        self.move_effector(pos_offset=np.asarray(self.pos_offset), vel_offset=np.asarray(self.vel_offset))

        # Run graphics commands to show/hide the plant if the visibility has changed
        self.update_plant_visibility()
        self.task_data['plant_visible'] = self.plant_visible

        # Save plant status to HDF file
        plant_data = self.plant.get_data_to_save() # for cursor plant, save cursor position
        for key in plant_data:
            self.task_data[key] = plant_data[key] # task_data['cursor']

        # Update the trial index
        self.task_data['trial'] = self.calc_trial_num()
        self.task_data['gen_idx'] = self.gen_index
        # print(self.task_data['gen_idx'])
        
        # Save the target position at each cycle. 
        if self.trial_timed_out:
             self.task_data['current_target'] = [np.nan,np.nan,np.nan]
             self.task_data['current_disturbance'] = [np.nan,np.nan,np.nan]
             self.task_data['current_target_validate'] = self.target.get_position() # default VirtualCircularTarget position is [0,0,0]
        else:
            self.task_data['current_target'] = self.targs[self.frame_index+self.lookahead]
            self.task_data['current_disturbance'] = self.disturbance_path[self.frame_index]
            self.task_data['current_target_validate'] = self.target.get_position()

        super()._cycle()

    def move_effector(self):
        '''Move the end effector, if a robot or similar is being controlled'''
        pass

    def run(self):
        '''
        See experiment.Experiment.run for trajectorydocumentation.
        '''
        # Fire up the plant. For virtual/simulation plants, this does little/nothing.
        self.plant.start()
        
        # Include some cleanup in case the parent class has errors
        try:
            super().run()
        finally:
            self.plant.stop()

    ##### HELPER AND UPDATE FUNCTIONS ####
    def update_plant_visibility(self):
        ''' Update plant visibility'''
        if self.plant_visible != self.plant_vis_prev:
            self.plant_vis_prev = self.plant_visible
            self.plant.set_visibility(self.plant_visible)

    def update_frame(self):
        self.target.move_to_position(np.array([0,0,self.targs[self.frame_index+self.lookahead][2]])) # xzy
        self.trajectory.move_to_position(np.array([-self.frame_index/3,0,0])) # same update constant works for 60 and 120 hz
        self.target.show()
        self.trajectory.show()
        self.frame_index +=1

    def setup_start_wait(self):
        if self.calc_trial_num() == 0:
            # Instantiate the targets here so they don't show up in any states that might come before "wait" 
            for model in self.target.graphics_models:
                self.add_model(model)
                self.target.hide()
                
            for model in self.trajectory.graphics_models:
                self.add_model(model)
                self.trajectory.hide()

            for model in self.bar.graphics_models:
                self.add_model(model)
                self.bar.hide()

        # Allow 2d movement
        if not self.always_1d:
            self.limit1d = False

        # Set up for progress bar
        self.bar_width = 12        
        self.tracking_frame_index = 0
        
        # Set up the next trajectory
        next_trajectory = np.array(np.squeeze(self.targs)[:,2])
        next_trajectory[:self.lookahead] = next_trajectory[self.lookahead]

        if hasattr(self, 'trajectory'):
            for model in self.trajectory.graphics_models:
                self.remove_model(model)
            del self.trajectory

        self.trajectory = VirtualCableTarget(target_radius=self.trajectory_radius, target_color=target_colors[self.trajectory_color], trajectory=next_trajectory)

        for model in self.trajectory.graphics_models:
                self.add_model(model)

        self.target.hide()
        self.trajectory.hide()   

    def setup_screen_reset(self):
        # Hide target and trajectory
        self.target.hide()
        self.target.reset()
        self.trajectory.hide()
        self.trajectory.reset()
        self.bar.hide()
        self.bar.reset()

    def setup_start_tracking_in(self):
        # Revert to settable trait
        self.limit1d = self.original_limit1d
        # Cue successful tracking
        self.target.cue_trial_end_success()

    def setup_start_tracking_out(self):
        # Reset target color
        self.target.reset()

    def setup_while_tracking(self):
        # Add disturbance
        cursor_pos = self.plant.get_endpoint_pos()
        if self.disturbance_trial == True:
            if self.velocity_control:
                # TODO check manualcontrolmixin for how to implement velocity control
                self.vel_offset = (cursor_pos + self.disturbance_path[self.frame_index])*1/self.fps
            else: 
                # position control
                self.pos_offset = self.disturbance_path[self.frame_index]

        # Move target and trajectory to next frame so it appears to be moving
        self.update_frame()

        # Check if the trial is over and there are no more target frames to display
        if self.frame_index+self.lookahead >= np.shape(self.targs)[0]:
            self.trial_timed_out = True

    #### TEST FUNCTIONS ####
    def _test_enter_target(self, time_in_state):
        '''
        Test if the cursor is inside the center target 
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.target.get_position())
        return d <= (self.target_radius - self.cursor_radius) or super()._test_enter_target(time_in_state) or (self.hold_time==0 and self.state=='trajectory')

    def _test_leave_target(self, time_in_state):
        '''
        Test if the cursor is outside the center target
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.target.get_position())
        return d > (self.target_radius - self.cursor_radius) or super()._test_leave_target(time_in_state)

    #### STATE FUNCTIONS ####
    def _start_wait(self):
        super()._start_wait()
        self.setup_start_wait()
        # print('WAIT')

    def _while_wait(self):
        super()._while_wait()

    def _start_wait_retry(self):
        super()._start_wait_retry()
        self.setup_start_wait()
        # print('WAIT RETRY')

    def _while_wait_retry(self):
        super()._while_wait_retry()

    def _start_trajectory(self):
        super()._start_trajectory()
        if self.frame_index == 0:
            self.target.move_to_position(np.array([0,0,self.targs[self.frame_index+self.lookahead][2]])) # tablet screen x-axis ranges -19,19, center 0
            self.trajectory.move_to_position(np.array([0,0,0])) # tablet screen x-axis ranges 0,41.33333, center 22ish
            # print(self.target.get_position())
            # print(self.trajectory.get_position())

            self.target.show()
            self.trajectory.show()
            # print('SHOW TRAJ')
            self.sync_event('TARGET_ON')

    def _while_trajectory(self):
        super()._while_trajectory()

    def _start_hold(self):
        super()._start_hold()
        # print('START HOLD')
        self.sync_event('TRIAL_START')
        # Cue successful tracking
        self.target.cue_trial_end_success()

    def _while_hold(self):
        super()._while_hold()

    def _start_tracking_in_ramp(self):
        super()._start_tracking_in_ramp()
        self.setup_start_tracking_in()
        # print('START TRACKING RAMP', self.ramp_counter[self.frame_index])
        self.sync_event('CURSOR_ENTER_TARGET', self.ramp_counter[self.frame_index])

    def _while_tracking_in_ramp(self):
        super()._while_tracking_in_ramp()
        self.setup_while_tracking()
        
    def _start_tracking_in(self):
        super()._start_tracking_in()
        self.setup_start_tracking_in()
        # print('START TRACKING')
        self.sync_event('CURSOR_ENTER_TARGET')

    def _while_tracking_in(self):
        super()._while_tracking_in()
        self.setup_while_tracking()
            
    def _start_tracking_out_ramp(self):
        super()._start_tracking_out_ramp()
        self.setup_start_tracking_out()
        # print('STOP TRACKING RAMP', self.ramp_counter[self.frame_index])
        self.sync_event('CURSOR_LEAVE_TARGET', self.ramp_counter[self.frame_index])

    def _while_tracking_out_ramp(self):
        super()._while_tracking_out_ramp()
        self.setup_while_tracking()

    def _start_tracking_out(self):
        super()._start_tracking_out()
        self.setup_start_tracking_out()
        # print('STOP TRACKING')
        self.sync_event('CURSOR_LEAVE_TARGET')

    def _while_tracking_out(self):
        super()._while_tracking_out()
        self.setup_while_tracking()

    def _start_timeout_penalty(self):
        super()._start_timeout_penalty()
        # print('START TIMEOUT')
        self.sync_event('TIMEOUT_PENALTY')
        self.setup_screen_reset()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_timeout_penalty(self):
        super()._while_timeout_penalty()

    def _end_timeout_penalty(self):
        super()._end_timeout_penalty()
        self.sync_event('TRIAL_END')
            
    def _start_hold_penalty(self):
        super()._start_hold_penalty()
        # print('START HOLD TIMEOUT')
        self.sync_event('HOLD_PENALTY')
        self.setup_screen_reset()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_hold_penalty(self):
        super()._while_hold_penalty()

    def _end_hold_penalty(self):
        super()._end_hold_penalty()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')

    def _start_tracking_out_penalty(self):
        super()._start_tracking_out_penalty()
        # print('START TRACKING TIMEOUT')
        self.sync_event('OTHER_PENALTY')

        # Cue failed trial
        self.target.cue_trial_end_failure()

        # skip to next generated trial using same freq set
        self.repeat_freq_set = True

    def _while_tracking_out_penalty(self):
        super()._while_tracking_out_penalty()

    def _end_tracking_out_penalty(self):
        super()._end_tracking_out_penalty()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')
        self.setup_screen_reset()

    def _start_reward(self):
        super()._start_reward()
        # print('REWARD')
        self.sync_event('REWARD')

        # Cue successful trial
        self.target.cue_trial_end_success()
        self.reward_frame_index = 0

        # use next generated trial using other freq set
        self.repeat_freq_set = False

    def _while_reward(self):
        super()._while_reward()

    def _end_reward(self):
        super()._end_reward()
        # print('TRIAL END')
        self.sync_event('TRIAL_END')
        self.setup_screen_reset()

    def _start_pause(self):
        super()._start_pause()
        # print('START PAUSE')
        self.sync_event('PAUSE_START') # self.sync_event('PAUSE_START') - this will overwrite TRIAL_END whenever you pause during an end state
        self.setup_screen_reset()

    def _while_pause(self):
        super()._while_pause()

    def _end_pause(self):
        super()._end_pause()
        # print('END PAUSE')
        self.sync_event('PAUSE_END')

    @staticmethod
    def calc_sum_of_sines(times, frequencies, amplitudes, phase_shifts):
        '''
        Generates the trajectories for the experiment
        '''
        t = times
        t = np.asarray(t).copy(); t.shape = (t.size,1)

        f = frequencies
        f = f.copy()
        f = np.reshape(f, (1,f.size))

        a = amplitudes
        a = a.copy()
        a = np.reshape(a, (1,a.size))

        p = phase_shifts        
        p = p.copy()
        p = np.reshape(p, (1,p.size))

        assert f.shape == a.shape == p.shape,"Shape of frequencies, amplitudes, and phase shifts must match"

        o = np.ones(t.shape)
        trajectory = np.sum(np.dot(o,a) * np.sin(2*np.pi*(np.dot(t,f) + np.dot(o,p))),axis=1)
        A = np.sum(np.dot(o,a)[0])
        
        return trajectory, A

    @staticmethod
    def calc_sum_of_sines_ramp(times, ramp, ramp_down, frequencies, amplitudes, phase_shifts):
            '''
            Adds a 1/t ramp up and ramp down at the start and end so the trajectories start and end at zero.
            '''
            t = times
            t = np.asarray(t).copy(); t.shape = (t.size,1)

            r = ramp
            rd = ramp_down

            trajectory, A = ScreenTargetTracking.calc_sum_of_sines(t, frequencies, amplitudes, phase_shifts)

            if r > 0:
                trajectory *= (( t*(t <= r)/r + (t > r) ).flatten())**2
                # trajectory *= (((t*(t <= r)/r) + ((t > r) & (t < (t[-1]-r))) + ((t[-1]-t)*(t >= (t[-1]-r))/r)).flatten())**2

            if rd > 0:
                trajectory *= (( (t < (t[-1]-rd)) + ((t[-1]-t)*(t >= (t[-1]-rd))/rd) ).flatten())**2

            return trajectory, A

    @staticmethod
    def generate_trajectories(num_trials=2, time_length=20, seed=40, sample_rate=120, base_period=20, ramp=0, ramp_down=0, num_primes=8):
        '''
        Sets up variables and uses prime numbers to call the above functions and generate then trajectories
        ramp is time length for preparatory lines
        '''
        np.random.seed(seed)
        hz = sample_rate # Hz -- sampling rate
        dt = 1/hz # sec -- sampling period

        T0 = base_period # sec -- base period
        w0 = 1./T0 # Hz -- base frequency

        r = ramp # "ramp up" duration (see sum_of_sines_ramp)
        rd = ramp_down # "ramp down" duration (see sum_of_sines_ramp)
        P = time_length/T0 # number of periods in signal
        T = P*T0+r+rd # sec -- signal duration
        dw = 1./T # Hz -- frequency resolution
        W = 1./dt/2 # Hz -- signal bandwidth

        full_primes = np.asarray([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 
            101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199])
        primes = full_primes[:num_primes]
        # primes_ind = np.where(full_primes <= T0)
        # primes = full_primes[primes_ind]

        f = primes*w0 # stimulated frequencies
        f_ref = f.copy()
        f_dis = f.copy()

        a = 1/(1+np.arange(f.size)) # amplitude
        a_ref = a.copy()
        a_dis = a.copy()

        o = np.random.rand(num_trials,primes.size) # phase offset
        o_ref = o.copy()
        o_dis = o.copy()*0.8

        t = np.arange(0,T,dt) # time samplesseed
        w = np.arange(0,W,dw) # frequency samples

        N = t.size # = T/dt -- number of samples
        
        # create trials dictionary
        trials = dict(
            id=np.arange(num_trials), times=np.tile(t,(num_trials,1)), ref=np.zeros((num_trials,N)), dis=np.zeros((num_trials,N))
            )

        # randomize order of first two trials to provide random starting point
        order = np.random.choice([0,1])
        if order == 0:
            trial_order = [(1,'E','O'),(1,'O','E')]
        elif order == 1:
            trial_order = [(1,'O','E'),(1,'E','O')]

        # generate reference and disturbance trajectories for all trials
        for trial_id, (num_reps,ref_ind,dis_ind) in enumerate(trial_order*int(num_trials/2)):   
            if ref_ind == 'E': 
                sines_r = np.arange(len(primes))[0::2] # use even indices
            elif ref_ind == 'O': 
                sines_r = np.arange(len(primes))[1::2] # use odd indices
            else:
                sines_r = np.arange(len(primes))
            if dis_ind == 'E':
                sines_d = np.arange(len(primes))[0::2]
            elif dis_ind == 'O':
                sines_d = np.arange(len(primes))[1::2]
            else:
                sines_d = np.arange(len(primes))
            
            ref_trajectory, ref_A = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, rd, f_ref[sines_r], a_ref[sines_r], o_ref[trial_id][sines_r])
            dis_trajectory, dis_A = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, rd, f_dis[sines_d], a_dis[sines_d], o_dis[trial_id][sines_d])
            
            # normalized trajectories
            trials['ref'][trial_id] = ref_trajectory/ref_A   # previously, denominator was np.sum(a_ref)
            trials['dis'][trial_id] = dis_trajectory/dis_A   # previously, denominator was np.sum(a_dis)
            # print(trial_order, ref_A, dis_A)
        
        return trials, trial_order

    @staticmethod
    def generate_trajectory(primes, base_period, ramp = .0):
        '''
        Sets up variables and uses prime numbers to call the above functions and generate then trajectories
        ramp is time length for preparatory lines
        '''
        hz = 60 # Hz -- sampling rate
        dt = 1/hz # sec -- sampling period

        T0 = base_period # sec -- base period
        w0 = 1/T0 # Hz -- base frequency

        P = 1 # number of periods in signal
        T = P*T0 # sec -- signal duration
        r = ramp # "ramp" duration (see sum_of_sines_ramp)
        dw = 1/T # Hz -- frequency resolution
        W = dw*T/dt/2 # Hz -- signal bandwidth

        f = primes # stimulated frequencies
        a = 1/(1+np.arange(f.size)) # amplitude
        o = 2*np.pi*np.random.rand(primes.size) # phase offset

        t = np.arange(0,T,dt) # time samples
        w = np.arange(0,W,dw) # frequency samples

        N = t.size # = T/dt -- number of samples

        #trajectory = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, f, a, o/(2*np.pi))
        trajectory = ScreenTargetTracking.calc_sum_of_sines_ramp(t, r, f, a, o)
        normalized_trajectory = trajectory/np.sum(a)
        return normalized_trajectory
    
    ### Generator functions ####
    @staticmethod
    def tracking_target_chain(nblocks=1, ntrials=2, time_length=20, ramp=0, ramp_down=0, num_primes=8, seed=40, sample_rate=120, disturbance=True, boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories

        Parameters
        ----------
        nblocks : int
            The number of blocks in the session
        ntrials : int
            The number of trials in a block
        time_length : int
            The length of one trial in seconds
        seed : int
            The seed for the random generator
        sample_rate : int
            The sample rate of the generated trajectories
        ramp : float
            The length of ramp up into a trial in seconds
        disturbance : boolean
            Whether to add disturbance to the cursor (disturbance is generated regardless)
        boundaries: 4 element tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        idx : [nblocks*ntrials x 1] array of trial indices
        targs : [nblocks*ntrials x 1] array of 3D target coordinates
        disturbance : boolean
        dis_trajectory : [nblocks*ntrials x 1] array of 3D disturbance coordinates
        '''
        idx = 0
        base_period = 20
        for block_id in range(nblocks):                
            trials, trial_order = ScreenTargetTracking.generate_trajectories(
                num_trials=ntrials, time_length=time_length, seed=seed, sample_rate=sample_rate, base_period=base_period, ramp=ramp, ramp_down=ramp_down, num_primes=num_primes
                )
            for trial_id in range(ntrials):
                targs = []
                ref_trajectory = np.zeros((int((time_length+ramp+ramp_down)*sample_rate),3))
                dis_trajectory = ref_trajectory.copy()
                ref_trajectory[:,2] = trials['ref'][trial_id]
                dis_trajectory[:,2] = trials['dis'][trial_id] # scale will determine lower limit of target size for perfect tracking
                targs.append(ref_trajectory)
                yield idx, targs, disturbance, dis_trajectory, sample_rate, ramp, ramp_down
                idx += 1

    @staticmethod
    def tracking_target_debug(nblocks=1, ntrials=2, time_length=20, seed=40, sample_rate=60, ramp=0, disturbance=True, boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories for debugging
        '''
        idx = 0
        base_period = 20
        for block_id in range(nblocks):                
            trials, trial_order = ScreenTargetTracking.generate_trajectories(
                num_trials=ntrials, time_length=time_length, seed=seed, sample_rate=sample_rate, base_period=base_period, ramp=ramp
                )
            for trial_id in range(ntrials):
                # if trial_id==0:
                #     plt.figure(); plt.plot(trials['times'][trial_id],trials['ref'][trial_id])
                #     plt.plot(trials['times'][trial_id],trials['dis'][trial_id])
                #     plt.show()
                targs = []
                ref_trajectory = np.zeros((int((time_length+ramp)*sample_rate),3))
                dis_trajectory = ref_trajectory.copy()
                ref_trajectory[:,2] = trials['ref'][trial_id]
                dis_trajectory[:,2] = trials['dis'][trial_id] # scale will determine lower limit of target size for perfect tracking
                targs.append(ref_trajectory)
                yield idx, targs, disturbance, dis_trajectory, sample_rate, ramp, ramp
                idx += 1
    
    @staticmethod
    def tracking_target_training(nblocks=1, ntrials=2, time_length=5, frequencies = [1,.75], boundaries=(-10,10,-10,10)):
        '''
        Generates a sequence of 1D (z axis) target trajectories for training

        Parameters
        ----------
        nblocks : int
            The number of tracking trials in the sequence.
        ntrials : int
            The number trials in a block
        time_length : int
        frequencies: numpy.ndarray
            A list of frequencies used to generate the trajectories
        boundaries: 4 element Tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        [nblocks*ntrials x 1] array of tuples containing trial indices and [time_length*60 x 3] target coordinates
        '''
        idx = 0
        buffer_space_bef = int(60*1.3) # 1.3 seconds of straight line before trial #78 frames
        buffer_space_aft = int(60*1.5) # 1.5 seconds of straight line before trial #90 frames
        frames = int(np.round(time_length*60)) #300 frames when time_length=5
        y_primes_freq = np.array(frequencies)

        for i in range(nblocks):
            for j in range(ntrials):
                
                disturbance = False
                disturbance_path = np.zeros((frames+buffer_space_bef+buffer_space_aft,1))
                trajectory = np.zeros((frames+buffer_space_bef+buffer_space_aft,3))
                sum_of_sins_path = ScreenTargetTracking.generate_trajectory(y_primes_freq,time_length)
                pts = []
                trajectory[:,2] = 5*np.concatenate((sum_of_sins_path[0]*np.ones(buffer_space_bef),sum_of_sins_path,sum_of_sins_path[-1]*np.ones(buffer_space_aft)))
                pts.append(trajectory)
                yield idx, pts, disturbance, disturbance_path, None
                idx += 1


'''
Classes for BMI decoding using linear scaling. 
'''
import numpy as np
from ..bmi.bmi import Filter

class State(object):
    '''For compatibility with other BMI decoding implementations'''

    def __init__(self, mean, *args, **kwargs):
        self.mean = mean

class LinearScaleFilter(Filter):

    model_attrs = ['attr']
    attrs_to_pickle = ['attr', 'obs', 'unit_to_state']

    def __init__(self, n_counts, n_states, n_units, unit_to_state=None, smoothing_window=1, decoder_to_plant=20, reject_threshold=5):
        '''
        Constructor for LinearScaleFilter

        Parameters
        ----------
        n_counts : Number of observations to hold
            Range is computed over the whole observation matrix size (N, D) 
            where N is the number of observations and D is the number of units
        n_states : How many state variables are there
            For example, a one-dim decoder has one state variable
        n_units : Number of neural units
            Can be number of isolated spiking units or number of channels for lfp
        unit_to_state : Which units to assign to which states (default = None)
            Floating point matrix of size (S, D) where S is the number of 
            states and D is the number of units, assigning a weight to each pair
        smoothing_window : How many observations to average to smooth output (default = 1)
        decoder_to_plant : how big is the screen, basically (default = 20)
            Maps from normalized output (0,1) to plant coordinates
        reject_threshold : How many standard deviations to reject
            If the standard deviation of the latest observation is greater than this,
            the filter will reject the observation
            
        Returns
        -------
        LinearScaleFilter instance
        '''
        self.obs = np.zeros((n_counts, n_units))
        self.n_states = n_states
        self.n_units = n_units
        self.smoothing_window = smoothing_window
        self.unit_to_state = unit_to_state
        self.decoder_to_plant = decoder_to_plant
        if unit_to_state is None:
            # Generate a default mapping where one unit controls one state
            self.unit_to_state = np.identity(max(n_states, n_units))
            self.unit_to_state = np.resize(self.unit_to_state, (n_states, n_units))
        self.count = 0
        self.attr = dict(
            neural_mean = np.zeros(n_units),
            neural_std = np.ones(n_units),
            offset = np.zeros(n_units),
            scale = np.ones(n_units),
        )
        self.fixed = False
        self.reject_threshold = reject_threshold
        self._init_state()

    def get_mean(self):
        ''' Must return self.state.mean to maintain compatibility'''
        return np.array(self.state.mean).ravel()

    def __call__(self, obs, **kwargs):
        ''' 
        Update the state with the new neural observation

        Parameters
        ----------
        obs : Neural observations (1 x P*N)
            where P is the number of features per channel and N is the number of channels
        kwargs : For consistency with other BMI decoders

        Returns
        ----------
        '''     

        # Z-score neural data
        norm_obs = (np.squeeze(obs) - self.attr['neural_mean']) / self.attr['neural_std']

        # Reject if the standard deviation of the latest observation is too high
        norm_obs[abs(norm_obs) > self.reject_threshold] = np.nan

        # Add observation
        self._add_obs(norm_obs, **kwargs)
        if not self.fixed:
            self._update_scale_attr()
        self._update_state()

    def update_norm_attr(self, neural_mean=None, neural_std=None, offset=None, scale=None):
        ''' Public method to set mean and std attributes'''
        if neural_mean is not None:
            self.attr.update(neural_mean = neural_mean)
        if neural_std is not None:
            self.attr.update(neural_std = neural_std)
        if offset is not None:
            self.attr.update(offset = offset)
        if scale is not None:
            self.attr.update(scale = scale)

    def fix_norm_attr(self):
        ''' Stop fliter from self updating its attributes'''
        self.fixed = True

    def unfix_norm_attr(self):
        ''' Let fliter update its scale attributes'''
        self.fixed = False

    def _pickle_init(self):
        #self.fix_norm_attr()
        pass

    def _init_state(self):
        ''' Required by decoder'''
        out = self._scale()
        self.state = State(out)

    def _update_state(self):
        out = self._scale()
        self.state = State(out)

    def _add_obs(self, obs,**kwargs):
        ''' Add new observations to the observation matrix'''
        
        if np.any(np.isnan(obs)):
            return

        # Update observation matrix
        if self.count < len(self.obs): 
            self.count += 1
        self.obs[:-1, :] = self.obs[1:, :]
        self.obs[-1, :] = obs
        
    def _get_latest_obs(self):
        ''' Smooth observations by taking the average across the smoothing window'''
        if self.count == 0:
            mean = np.zeros(np.size(self.obs, axis=1))
        elif self.count < self.smoothing_window:
            mean = np.squeeze(np. nanmean(self.obs[-self.count:, :], axis=0))
        else:
            mean = np.squeeze(np.nanmean(self.obs[-self.smoothing_window:, :], axis=0))
        return mean
        
    def _scale(self):
        ''' Scale the (normalized) observations '''

        # Normalize windowed average to the current offset and scale
        x = (self._get_latest_obs() - self.attr['offset']) / self.attr['scale']
        
        # Arrange output according to map
        out = np.matmul(self.unit_to_state, x).reshape(-1,1) * self.decoder_to_plant
        return out

    def _update_scale_attr(self):
        ''' Update the normalization parameters'''

        # Normalize latest observation(s)
        mean = np.nanmedian(self.obs[-self.count:, :], axis=0)
        # range = max(1, np.amax(self.obs[-self.count:, :]) - np.amin(self.obs[-self.count:, :]))
        std = np.nanstd(self.obs[-self.count:, :], axis=0)
        std[std < 1e-6] = 1e-6 # Avoid divide by zero
        self.update_norm_attr(offset=mean, scale=std)


class PosVelState(State):
    ''' 
    Simple state with the ability to integrate velocity over time
    Only compatible with StateSpaceEndptVel2D state space matrices
    '''

    def __init__(self, vel_control, call_rate=60):
        self.vel_control = vel_control
        self.call_rate = call_rate
        self.mean = np.zeros((7,1))

    def update(self, mean):
        if self.vel_control:
            self.mean[3:6] = mean[3:6]

            # Add the velocity (units/s) to the position (units)
            self.mean[0:3] = self.mean[3:6] / self.call_rate + self.mean[0:3]
        else:
            self.mean = mean

class PosVelScaleFilter(LinearScaleFilter):
    ''' 
    Linear filter that holds a position and velocity state
    Only compatible with StateSpaceEndptVel2D state space matrices
    '''

    def __init__(self, vel_control, *args, **kwargs):
        self.call_rate = kwargs.pop('call_rate')
        self.vel_control = vel_control
        super(PosVelScaleFilter, self).__init__(*args, **kwargs)

    def _init_state(self):
        self.state = PosVelState(self.vel_control, self.call_rate)
        out = self._scale()
        self.state.update(out)        

    def _update_state(self):
        out = self._scale()
        self.state.update(out)  

def create_lindecoder(ssm, units, unit_to_state, decoder_to_plant=None, smoothing_window=1, vel_control=False, update_rate=0.1):
    from riglib.bmi import Decoder
    filt_counts = smoothing_window # only used for smoothing since we're fixing the gains
    filt = PosVelScaleFilter(vel_control, filt_counts, ssm.n_states, len(units), unit_to_state=unit_to_state, smoothing_window=smoothing_window, decoder_to_plant=decoder_to_plant, call_rate=1/update_rate)
    
    # calculate gains from training data
    # filt.update_norm_attr(neural_mean=0, neural_std=1, offset=mFR, scale=sdFR)
    
    decoder = Decoder(filt, units, ssm, binlen=update_rate, subbins=1)
    decoder.n_features = len(units)
    decoder.binlen = update_rate

    return decoder


from OpenGL.GL import *
from matplotlib import projections
from .fbo import FBO, FBOrender
from ..utils import look_at, orthographic, perspective
import numpy as np

class ShadowMapper(FBOrender):
    def __init__(self, *args, **kwargs):
        super(ShadowMapper, self).__init__(*args, **kwargs)
        self.size = (1024, 1024)
        
        # Create shadow map FBO
        self.shadow_map = FBO(["depth"], size=self.size)

        # Add shaders for shadow pass
        self.add_shader("shadow_pass_v", GL_VERTEX_SHADER, "shadow_pass.v.glsl")
        self.add_shader("shadow_pass_f", GL_FRAGMENT_SHADER, "shadow_pass.f.glsl")
        self.add_program("shadow_pass", ("shadow_pass_v", "shadow_pass_f"))

        self.add_shader("fsquad", GL_VERTEX_SHADER, "fsquad.v.glsl")
        self.add_shader("fsquad_frag", GL_FRAGMENT_SHADER, "fsquad.f.glsl")
        self.add_program("none", ("fsquad", "fsquad_frag"))

        # Calculate light space matrix
        light_target = [0, 0, 0]
        light_up = [0, 1, 0]
        near_plane = 25.0
        far_plane = 100.0
        light_projection = perspective(75, 1, near_plane, far_plane)
        light_view = look_at(-50*np.array(self.light_direction[:3])/np.linalg.norm(self.light_direction[:3]), light_target, light_up)
        self.light_space_matrix = np.dot(light_projection, light_view)


    def generate_shadow_map(self, root, **kwargs):
        # Save current viewport and framebuffer
        original_viewport = glGetIntegerv(GL_VIEWPORT)
        original_framebuffer = glGetIntegerv(GL_FRAMEBUFFER_BINDING)
        original_winding_order = glGetIntegerv(GL_FRONT_FACE)

        # Set viewport for shadow map
        glViewport(0, 0, self.size[0], self.size[1])
        glFrontFace(GL_CCW)
        glCullFace(GL_FRONT)

        # Render shadow map
        self.draw_to_fbo(self.shadow_map, root, shader="shadow_pass", apply_default=True, 
                         light_space_matrix=self.light_space_matrix, **kwargs)

        # Restore original viewport and framebuffer
        glFrontFace(original_winding_order)
        glViewport(*original_viewport)
        glBindFramebuffer(GL_FRAMEBUFFER, original_framebuffer)
        glCullFace(GL_BACK)

        return self.shadow_map['depth'], self.light_space_matrix
    
    def draw(self, root, **kwargs):
        # Generate shadow map
        shadow_map, light_space_matrix = self.generate_shadow_map(
            root, **kwargs
        )
        # self.draw_fsquad("none", tex=shadow_map)
        # return
        # super().draw(root, p_matrix=light_space_matrix, modelview=np.eye(4))
        # return
    
        # Draw the scene with shadows
        super().draw(root, light_space_matrix=light_space_matrix,
                           shadow_map=shadow_map,
                           **kwargs)


'''
This module implements IK functions for running inverse kinematics.
Current, only a two-joint system can be modelled.
'''

import numpy as np

from .xfm import Quaternion
from .models import Group
from .primitives import Cylinder, Sphere, Cone
from ..bmi import robot_arms

pi = np.pi

joint_angles_dtype = [('sh_pflex', np.float64), ('sh_pabd', np.float64), ('sh_prot', np.float64), ('el_pflex', np.float64), ('el_psup', np.float64)]
joint_vel_dtype = [('sh_vflex', np.float64), ('sh_vabd', np.float64), ('sh_vrot', np.float64), ('el_vflex', np.float64), ('el_vsup', np.float64)]

def inv_kin_2D(pos, l_upperarm, l_forearm, vel=None):
    '''
    Inverse kinematics for a 2D arm. This function returns all 5 angles required
    to specify the pose of the exoskeleton (see riglib.bmi.train for the 
    definitions of these angles). This pose is constrained to the x-z plane
    by forcing shoulder flexion/extension, elbow rotation and supination/pronation
    to always be 0. 
    '''
    if np.ndim(pos) == 1:
        pos = pos.reshape(1,-1)

    # require the y-coordinate to be 0, i.e. flat on the screen
    x, y, z = pos[:,0], pos[:,1], pos[:,2]
    assert np.all(y == 0)

    if vel is not None:
        if np.ndim(vel) == 1:
            vel = vel.reshape(1,-1)
        assert pos.shape == vel.shape
        vx, vy, vz = vel[:,0], vel[:,1], vel[:,2]
        assert np.all(vy == 0)

    L = np.sqrt(x**2 + z**2)
    cos_el_pflex = (L**2 - l_forearm**2 - l_upperarm**2) / (2*l_forearm*l_upperarm)

    cos_el_pflex[ (cos_el_pflex > 1) & (cos_el_pflex < 1 + 1e-9)] = 1
    el_pflex = np.arccos(cos_el_pflex)

    sh_pabd = np.arctan2(z, x) - np.arcsin(l_forearm * np.sin(np.pi - el_pflex) / L)
    angles = np.zeros(len(pos), dtype=joint_angles_dtype)
    angles['sh_pabd'] = sh_pabd
    angles['el_pflex'] = el_pflex
    if np.any(np.isnan(angles['el_pflex'])) or np.any(np.isnan(angles['sh_pabd'])):
        pass

    if vel is not None:
        joint_vel = np.zeros(len(pos), dtype=joint_vel_dtype)
        
        # Calculate the jacobian
        for k, angle in enumerate(angles):
            s1 = np.sin(angle['sh_pabd'])
            s2 = np.sin(angle['sh_pabd'] + angle['el_pflex'])
            c1 = np.cos(angle['sh_pabd'])
            c2 = np.cos(angle['sh_pabd'] + angle['el_pflex'])           
            J = np.array([[-l_upperarm*s1-l_forearm*s2, -l_forearm*s2 ],
                          [ l_upperarm*c1+l_forearm*c2,  l_forearm*c2 ]])
            J_inv = np.linalg.inv(J)

            joint_vel_mat = np.dot(J_inv, vel[k, [0,2]])
            joint_vel[k]['sh_vabd'], joint_vel[k]['el_vflex'] = joint_vel_mat.ravel()

        return angles, joint_vel
    else:
        return angles    

def make_list(value, num_joints):
    '''
    Helper function to allow joint/link properties of the chain to be specified
    as one value for all joints/links or as separate values for each
    '''
    if isinstance(value, list) and len(value) == num_joints:
        return value
    else:
        return [value] * num_joints

arm_color = (181/256., 116/256., 96/256., 1)
arm_radius = 0.6

class Plant(object):
    def __init__(self, *args, **kwargs):
        super(Plant, self).__init__(*args, **kwargs)

    def drive(self, decoder):
        self.set_intrinsic_coordinates(decoder['q'])
        intrinsic_coords = self.get_intrinsic_coordinates()
        if not np.any(np.isnan(intrinsic_coords)):
            decoder['q'] = self.get_intrinsic_coordinates()        


class CursorPlant(Plant):
    def __init__(self, endpt_bounds=None, **kwargs):
        self.endpt_bounds = endpt_bounds
        self.position = np.array([0., 0., 0.]) #np.zeros(3)

    def get_endpoint_pos(self):
        return self.position

    def set_endpoint_pos(self, pt, **kwargs):
        self.position = pt

    def get_intrinsic_coordinates(self):
        return self.position

    def set_intrinsic_coordinates(self, pt):
        self.position = pt

    def drive(self, decoder):
        pos = decoder['q'].copy()
        vel = decoder['qdot'].copy()
        
        if self.endpt_bounds is not None:
            if pos[0] < self.endpt_bounds[0]: 
                pos[0] = self.endpt_bounds[0]
                #vel[0] = 0
            if pos[0] > self.endpt_bounds[1]: 
                pos[0] = self.endpt_bounds[1]
                #vel[0] = 0

            if pos[1] < self.endpt_bounds[2]: 
                pos[1] = self.endpt_bounds[2]
                #vel[1] = 0
            if pos[1] > self.endpt_bounds[3]: 
                pos[1] = self.endpt_bounds[3]
                #vel[1] = 0

            if pos[2] < self.endpt_bounds[4]: 
                pos[2] = self.endpt_bounds[4]
                #vel[2] = 0
            if pos[2] > self.endpt_bounds[5]: 
                pos[2] = self.endpt_bounds[5]
                #vel[2] = 0
        
        decoder['q'] = pos
        decoder['qdot'] = vel
        super(CursorPlant, self).drive(decoder)


class RobotArmGen2D(Plant, Group):
    def __init__(self, link_radii=arm_radius, joint_radii=arm_radius, link_lengths=[15,15,5,5], joint_colors=arm_color,
        link_colors=arm_color, base_loc=np.array([2., 0., -15]), **kwargs):
        '''
        Instantiate the graphics and the virtual arm for a planar kinematic chain
        '''
        num_joints = len(link_lengths)
        self.num_joints = num_joints

        self.link_radii = make_list(link_radii, num_joints)
        self.joint_radii = make_list(joint_radii, num_joints)
        self.link_lengths = make_list(link_lengths, num_joints)
        self.joint_colors = make_list(joint_colors, num_joints)
        self.link_colors = make_list(link_colors, num_joints)

        self.curr_vecs = np.zeros([num_joints, 3]) #rows go from proximal to distal links

        # set initial vecs to correct orientations (arm starts out vertical)
        self.curr_vecs[0,2] = self.link_lengths[0]
        self.curr_vecs[1:,0] = self.link_lengths[1:]

        # Create links
        self.links = []

        for i in range(self.num_joints):
            joint = Sphere(radius=self.joint_radii[i], color=self.joint_colors[i])

            # The most distal link gets a tapered cylinder (for purely stylistic reasons)
            if i < self.num_joints - 1:
                link = Cylinder(radius=self.link_radii[i], height=self.link_lengths[i], color=self.link_colors[i])
            else:
                link = Cone(radius1=self.link_radii[-1], radius2=self.link_radii[-1]/2, height=self.link_lengths[-1], color=self.link_colors[-1])
            link_i = Group((link, joint))
            self.links.append(link_i)

        link_offsets = [0] + self.link_lengths[:-1]
        self.link_groups = [None]*self.num_joints
        for i in range(self.num_joints)[::-1]:
            if i == self.num_joints-1:
                self.link_groups[i] = self.links[i]
            else:
                self.link_groups[i] = Group([self.links[i], self.link_groups[i+1]])

            self.link_groups[i].translate(0, 0, link_offsets[i])

        # Call the parent constructor
        super(RobotArmGen2D, self).__init__([self.link_groups[0]], **kwargs)

        # Instantiate the kinematic chain object
        if self.num_joints == 2:
            self.kin_chain = robot_arms.PlanarXZKinematicChain(link_lengths)
            self.kin_chain.joint_limits = [(-pi,pi), (-pi,0)]
        else:
            self.kin_chain = robot_arms.PlanarXZKinematicChain(link_lengths)

            # TODO the code below is (obviously) specific to a 4-joint chain
            self.kin_chain.joint_limits = [(-pi,pi), (-pi,0), (-pi/2,pi/2), (-pi/2, 10*pi/180)]

        self.base_loc = base_loc
        self.translate(*self.base_loc, reset=True)

    def _update_links(self):
        for i in range(0, self.num_joints):
            # Rotate each joint to the vector specified by the corresponding row in self.curr_vecs
            # Annoyingly, the baseline orientation of the first group is always different from the 
            # more distal attachments, so the rotations have to be found relative to the orientation 
            # established at instantiation time.
            if i == 0:
                baseline_orientation = (0, 0, 1)
            else:
                baseline_orientation = (1, 0, 0)

            # Find the normalized quaternion that represents the desired joint rotation
            self.link_groups[i].xfm.rotate = Quaternion.rotate_vecs(baseline_orientation, self.curr_vecs[i]).norm()

            # Recompute any cached transformations after the change
            self.link_groups[i]._recache_xfm()

    def get_endpoint_pos(self):
        '''
        Returns the current position of the non-anchored end of the arm.
        '''
        relangs = np.arctan2(self.curr_vecs[:,2], self.curr_vecs[:,0])
        return self.perform_fk(relangs) + self.base_loc

    def perform_fk(self, angs):
        absvecs = np.zeros(self.curr_vecs.shape)
        for i in range(self.num_joints):
            absvecs[i] = self.link_lengths[i]*np.array([np.cos(np.sum(angs[:i+1])), 0, np.sin(np.sum(angs[:i+1]))])
        return np.sum(absvecs,axis=0)

    def set_endpoint_pos(self, pos, **kwargs):
        '''
        Positions the arm according to specified endpoint position. 
        '''
        if pos is not None:
            # Run the inverse kinematics
            angles = self.perform_ik(pos, **kwargs)

            # Update the joint configuration    
            self.set_intrinsic_coordinates(angles)

    def perform_ik(self, pos, **kwargs):
        angles = self.kin_chain.inverse_kinematics(pos - self.base_loc, q_start=-self.get_intrinsic_coordinates(), verbose=False, eps=0.008, **kwargs)
        # print self.kin_chain.endpoint_pos(angles)

        # Negate the angles. The convention in the robotics library is 
        # inverted, i.e. in the robotics library, positive is clockwise 
        # rotation whereas here CCW rotation is positive. 
        angles = -angles        
        return angles

    def calc_joint_angles(self, vecs):
        return np.arctan2(vecs[:,2], vecs[:,0])

    def get_intrinsic_coordinates(self):
        '''
        Returns the joint angles of the arm in radians
        '''
        
        return self.calc_joint_angles(self.curr_vecs)
        
    def set_intrinsic_coordinates(self,theta):
        '''
        Set the joint by specifying the angle in radians. Theta is a list of angles. If an element of theta = NaN, angle should remain the same.
        '''
        for i in range(self.num_joints):
            if theta[i] is not None and ~np.isnan(theta[i]):
                self.curr_vecs[i] = self.link_lengths[i]*np.array([np.cos(theta[i]), 0, np.sin(theta[i])])
                
        self._update_links()


class RobotArmGen3D(Plant, Group):
    def __init__(self, link_radii=arm_radius, joint_radii=arm_radius, link_lengths=[15,15,5,5], joint_colors=arm_color,
        link_colors=arm_color, base_loc=np.array([2., 0., -15]), **kwargs):
        '''
        Instantiate the graphics and the virtual arm for a kinematic chain
        '''
        num_joints = 2
        self.num_joints = 2

        self.link_radii = make_list(link_radii, num_joints)
        self.joint_radii = make_list(joint_radii, num_joints)
        self.link_lengths = make_list(link_lengths, num_joints)
        self.joint_colors = make_list(joint_colors, num_joints)
        self.link_colors = make_list(link_colors, num_joints)

        self.curr_vecs = np.zeros([num_joints, 3]) #rows go from proximal to distal links

        # set initial vecs to correct orientations (arm starts out vertical)
        self.curr_vecs[0,2] = self.link_lengths[0]
        self.curr_vecs[1:,0] = self.link_lengths[1:]

        # Create links
        self.links = []

        for i in range(self.num_joints):
            joint = Sphere(radius=self.joint_radii[i], color=self.joint_colors[i])

            # The most distal link gets a tapered cylinder (for purely stylistic reasons)
            if i < self.num_joints - 1:
                link = Cylinder(radius=self.link_radii[i], height=self.link_lengths[i], color=self.link_colors[i])
            else:
                link = Cone(radius1=self.link_radii[-1], radius2=self.link_radii[-1]/2, height=self.link_lengths[-1], color=self.link_colors[-1])
            link_i = Group((link, joint))
            self.links.append(link_i)

        link_offsets = [0] + self.link_lengths[:-1]
        self.link_groups = [None]*self.num_joints
        for i in range(self.num_joints)[::-1]:
            if i == self.num_joints-1:
                self.link_groups[i] = self.links[i]
            else:
                self.link_groups[i] = Group([self.links[i], self.link_groups[i+1]])

            self.link_groups[i].translate(0, 0, link_offsets[i])

        # Call the parent constructor
        super(RobotArmGen3D, self).__init__([self.link_groups[0]], **kwargs)

        # Instantiate the kinematic chain object
        if self.num_joints == 2:
            self.kin_chain = robot_arms.PlanarXZKinematicChain(link_lengths)
            self.kin_chain.joint_limits = [(-pi,pi), (-pi,0)]
        else:
            self.kin_chain = robot_arms.PlanarXZKinematicChain(link_lengths)

            # TODO the code below is (obviously) specific to a 4-joint chain
            self.kin_chain.joint_limits = [(-pi,pi), (-pi,0), (-pi/2,pi/2), (-pi/2, 10*pi/180)]

        self.base_loc = base_loc
        self.translate(*self.base_loc, reset=True)

    def _update_links(self):
        for i in range(0, self.num_joints):
            # Rotate each joint to the vector specified by the corresponding row in self.curr_vecs
            # Annoyingly, the baseline orientation of the first group is always different from the 
            # more distal attachments, so the rotations have to be found relative to the orientation 
            # established at instantiation time.
            if i == 0:
                baseline_orientation = (0, 0, 1)
            else:
                baseline_orientation = (1, 0, 0)

            # Find the normalized quaternion that represents the desired joint rotation
            self.link_groups[i].xfm.rotate = Quaternion.rotate_vecs(baseline_orientation, self.curr_vecs[i]).norm()

            # Recompute any cached transformations after the change
            self.link_groups[i]._recache_xfm()

    def get_endpoint_pos(self):
        '''
        Returns the current position of the non-anchored end of the arm.
        '''
        relangs_xz = np.arctan2(self.curr_vecs[:,2], self.curr_vecs[:,0])
        relangs_xy = np.arctan2(self.curr_vecs[:,1], self.curr_vecs[:,0])
        return self.perform_fk(relangs_xz, relangs_xy) + self.base_loc

    def perform_fk(self, angs_xz, angs_xy):
        absvecs = np.zeros(self.curr_vecs.shape)
        for i in range(self.num_joints):
            absvecs[i] = self.link_lengths[i]*np.array([np.cos(np.sum(angs_xz[:i+1])), np.sin(np.sum(angs_xy[:i+1])), np.sin(np.sum(angs_xz[:i+1]))])
        return np.sum(absvecs,axis=0)

    def set_endpoint_pos(self, pos, **kwargs):
        '''
        Positions the arm according to specified endpoint position. 
        '''
        if pos is not None:
            # Run the inverse kinematics
            angles = self.perform_ik(pos, **kwargs)

            # Update the joint configuration    
            self.set_intrinsic_coordinates(angles)


    def perform_ik(self, pos, **kwargs):
        angles = self.kin_chain.inverse_kinematics(pos - self.base_loc, q_start=-self.get_intrinsic_coordinates(), verbose=False, eps=0.008, **kwargs)
        # print self.kin_chain.endpoint_pos(angles)

        # Negate the angles. The convention in the robotics library is 
        # inverted, i.e. in the robotics library, positive is clockwise 
        # rotation whereas here CCW rotation is positive. 
        angles = -angles




        '''Sets the endpoint coordinate for the two-joint system'''
        #Make sure the target is actually achievable
        if np.linalg.norm(pos) > self.tlen:
            self.upperarm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), target).norm()
            self.forearm.xfm.rotate = Quaternion()
        else:
            elbow = np.array(self._midpos(target))
            
            #rotate the upperarm to the elbow
            self.upperarm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), elbow).norm()

            #this broke my mind for 2 hours at least, so I cheated
            #Rotate first to (0,0,1), then rotate to the target-elbow
            self.forearm.xfm.rotate = (Quaternion.rotate_vecs(elbow, (0,0,1)) *
                Quaternion.rotate_vecs((0,0,1), target-elbow)).norm()
        
        self.upperarm._recache_xfm()
        # print self.upperarm.xfm.rotate
        self.curr_vecs[0] = self.lengths[0]*self.upperarm.xfm.rotate.quat[1:]
        self.curr_vecs[1] = self.lengths[1]*self.forearm.xfm.rotate.quat[1:]
        print(self.forearm.xfm)
        print(self.upperarm.xfm)
        # raise NotImplementedError("update curr_vecs!")

              
        return angles

    def calc_joint_angles(self, vecs):
        return np.arctan2(vecs[:,2], vecs[:,0]), np.arctan2(vecs[:,1], vecs[:,0])

    def get_intrinsic_coordinates(self):
        '''
        Returns the joint angles of the arm in radians
        '''
        
        return self.calc_joint_angles(self.curr_vecs)
        
    def set_intrinsic_coordinates(self,theta_xz, theta_xy):
        '''
        Set the joint by specifying the angle in radians. Theta is a list of angles. If an element of theta = NaN, angle should remain the same.
        '''
        for i in range(self.num_joints):
            if theta_xz[i] is not None and ~np.isnan(theta_xz[i]) and theta_xy[i] is not None and ~np.isnan(theta_xy[i]):
                self.curr_vecs[i] = self.link_lengths[i]*np.array([np.cos(theta_xz[i]), np.sin(theta_xy[i]), np.sin(theta_xz[i])])
                
        self._update_links()

    def _midpos(self, target):
            m, n = self.link_lengths
            x, y, z = target

            #this heinous equation brought to you by Wolfram Alpha
            #it is ONE of the solutions to this system of equations:
            # a^2 + b^2 + (z/2)^2 = m^2
            # (x-a)^2 + (y-b)^2 + (z/2)^2 = n^2
            if x > 0:
                a = (m**2*x**2+y*np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*x**2+x**4+x**2*y**2)/(2*x*(x**2+y**2))
                b = (m**2*y-np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*y+x**2*y+y**3)/(2*(x**2+y**2))
            else:
                a = (m**2*x**2-y*np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*x**2+x**4+x**2*y**2)/(2*x*(x**2+y**2))
                b = (m**2*y+np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*y+x**2*y+y**3)/(2*(x**2+y**2))
            return a, b, z/2


class RobotArm2J2D(RobotArmGen2D):
    def drive(self, decoder):
        raise NotImplementedError("deal with the state bounding stuff!")
        # elif self.decoder.ssm == train.joint_2D_state_space:
        #     self.set_arm_joints(self.decoder['sh_pabd', 'el_pflex'])

        #     # Force the arm to a joint configuration where the cursor is on-screen
        #     pos = self.get_arm_endpoint()
        #     pos = self.apply_cursor_bounds(pos)
        #     self.set_arm_endpoint(pos)

        #     # Reset the decoder state to match the joint configuration of the arm
        #     joint_pos = self.get_arm_joints()
        #     self.decoder['sh_pabd', 'el_pflex'] = joint_pos

class TwoJoint(object):
    '''
    Models a two-joint IK system (arm, leg, etc). Constrains the system by 
    always having middle joint halfway between the origin and target
    '''
    def __init__(self, origin_bone, target_bone, lengths=(20,20)):
        '''Takes two Model objects for the "upperarm" bone and the "lowerarm" bone.
        Assumes the models start at origin, with vector to (0,0,1) for bones'''
        self.upperarm = origin_bone
        self.forearm = target_bone
        self.lengths = lengths
        self.tlen = lengths[0] + lengths[1]
        self.curr_vecs = np.zeros([2,3])
        self.curr_angles = np.zeros(2)

    def _midpos(self, target):
        m, n = self.lengths
        x, y, z = target

        #this heinous equation brought to you by Wolfram Alpha
        #it is ONE of the solutions to this system of equations:
        # a^2 + b^2 + (z/2)^2 = m^2
        # (x-a)^2 + (y-b)^2 + (z/2)^2 = n^2
        if x > 0:
            a = (m**2*x**2+y*np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*x**2+x**4+x**2*y**2)/(2*x*(x**2+y**2))
            b = (m**2*y-np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*y+x**2*y+y**3)/(2*(x**2+y**2))
        else:
            a = (m**2*x**2-y*np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*x**2+x**4+x**2*y**2)/(2*x*(x**2+y**2))
            b = (m**2*y+np.sqrt(-x**2*(m**4-2*m**2*n**2-2*m**2*x**2-2*m**2*y**2+n**4-2*n**2*x**2-2*n**2*y**2+x**4+2*x**2*y**2+x**2*z**2+y**4+y**2*z**2))-n**2*y+x**2*y+y**3)/(2*(x**2+y**2))
        return a, b, z/2

    def set_endpoint_3D(self, target):
        '''Sets the endpoint coordinate for the two-joint system'''
        #Make sure the target is actually achievable
        if np.linalg.norm(target) > self.tlen:
            self.upperarm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), target).norm()
            self.forearm.xfm.rotate = Quaternion()
        else:
            elbow = np.array(self._midpos(target))
            
            #rotate the upperarm to the elbow
            self.upperarm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), elbow).norm()

            #this broke my mind for 2 hours at least, so I cheated
            #Rotate first to (0,0,1), then rotate to the target-elbow
            self.forearm.xfm.rotate = (Quaternion.rotate_vecs(elbow, (0,0,1)) *
                Quaternion.rotate_vecs((0,0,1), target-elbow)).norm()
        
        self.upperarm._recache_xfm()
        # print self.upperarm.xfm.rotate
        upperarm_affine_xform = self.upperarm.xfm.rotate.to_mat()
        forearm_affine_xform = (self.upperarm.xfm * self.forearm.xfm).rotate.to_mat()
        # print np.dot(upperarm_affine_xform, np.array([0., 0, self.lengths[0], 1]))
        self.curr_vecs[0] = np.dot(upperarm_affine_xform, np.array([0., 0, self.lengths[0], 1]))[:-1]#self.lengths[0]*self.upperarm.xfm.rotate.quat[1:]
        self.curr_vecs[1] = np.dot(forearm_affine_xform, np.array([0, 0, self.lengths[1], 1]))[:-1]#self.lengths[1]*self.forearm.xfm.rotate.quat[1:]
        # print self.forearm.xfm
        # print self.upperarm.xfm
        # raise NotImplementedError("update curr_vecs!")

    def set_endpoint_2D(self, target):
        ''' Given an endpoint coordinate in the x-z plane, solves for joint positions in that plane via inverse kinematics'''
        pass

    def set_joints_2D(self, shoulder_angle, elbow_angle):
        ''' Given angles for shoulder and elbow in a plane, set joint positions. Shoulder angle is in fixed
        frame of reference where 0 is horizontal pointing to the right (left if viewing on screen without mirror), pi/2
        is vertical pointing up, pi is horizontal pointing to the left. Elbow angle is relative to upper arm vector, where
        0 is fully extended, pi/2 is a right angle to upper arm pointing left, and pi is fully overlapping with upper
        arm.'''

        elbow_angle_mod = elbow_angle + np.pi/2

        #if shoulder_angle>np.pi: shoulder_angle = np.pi
        #if shoulder_angle<0.0: shoulder_angle = 0.0
        #if elbow_angle>np.pi: elbow_angle = np.pi
        #if elbow_angle<0: elbow_angle = 0

        # Find upper arm vector
        xs = self.lengths[0]*np.cos(shoulder_angle)
        ys = 0.0
        zs = self.lengths[0]*np.sin(shoulder_angle)
        self.curr_vecs[0,:] = np.array([xs, ys, zs])
        self.upperarm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), (xs,0,zs)).norm()

        # Find forearm vector (relative to upper arm)
        xe = self.lengths[1]*np.cos(elbow_angle_mod)
        ye = 0.0
        ze = self.lengths[1]*np.sin(elbow_angle_mod)
        # Find absolute vector
        xe2 = self.lengths[1]*np.cos(shoulder_angle+elbow_angle)
        ye2 = 0.0
        ze2 = self.lengths[1]*np.sin(shoulder_angle+elbow_angle)
        self.curr_vecs[1,:] = np.array([xe2, ye2, ze2])
        self.forearm.xfm.rotate = Quaternion.rotate_vecs((0,0,1), (xe,0,ze)).norm()

        self.curr_angles[0] = shoulder_angle
        self.curr_angles[1] = elbow_angle_mod

        self.upperarm._recache_xfm()


        # cursor_color = (.5,0,.5,1)
        # cursor_radius = 0.4
        # self.endpt_cursor = Sphere(radius=cursor_radius, color=cursor_color)
        # self.endpt_cursor.translate(0, 0, lengths[1])
        # self.forearm = Group([
        #     Cylinder(radius=link_radii[1], height=lengths[1], color=link_colors[1]), 
        #     self.endpt_cursor])
        # self.forearm.translate(0,0,lengths[0])
        
        # self.upperarm = Group([
        #     Cylinder(radius=link_radii[0], height=lengths[0],color=link_colors[0]), 
        #     Sphere(radius=ball_radii[0],color=ball_colors[0]).translate(0, 0, lengths[0]),
        #     self.forearm])
        # self.system = TwoJoint(self.upperarm, self.forearm, lengths = (self.lengths))

class RobotArm(Plant, Group):
    def __init__(self, link_radii=(.2, .2), ball_radii=(.5,.5),lengths=(20, 20), ball_colors = ((1,1,1,1),(1,1,1,1)),\
        link_colors = ((1,1,1,1), (1,1,1,1)), base_loc=np.array([2., 0., -10.]), **kwargs):
        self.link_radii = link_radii
        self.ball_radii = ball_radii
        self.lengths = lengths

        self.endpt_cursor = Sphere(radius=ball_radii[1], color=(1, 0, 1, 1)) #ball_colors[1])
        self.forearm = Group([
            Cylinder(radius=link_radii[1], height=lengths[1], color=link_colors[1]), 
            self.endpt_cursor.translate(0, 0, lengths[1])]).translate(0,0,lengths[0])
        self.upperarm = Group([
            Cylinder(radius=link_radii[0], height=lengths[0],color=link_colors[0]), 
            Sphere(radius=ball_radii[0],color=ball_colors[0]).translate(0, 0, lengths[0]),
            self.forearm])
        self.system = TwoJoint(self.upperarm, self.forearm, lengths = (self.lengths))
        super(RobotArm, self).__init__([self.upperarm], **kwargs)

        self.num_links = len(link_radii)
        self.num_joints = 3 # abstract joints. this system is fully characterized by the endpoint position since the elbow angle is determined by IK

        self.base_loc = base_loc

        self.translate(*self.base_loc, reset=True)

    def get_endpoint_pos(self):
        # print 'curr_vecs', self.system.curr_vecs
        # print
        return np.sum(self.system.curr_vecs, axis=0) + self.base_loc

    def set_endpoint_pos(self, pos, **kwargs):
        self.system.set_endpoint_3D(pos - self.base_loc)

    def get_intrinsic_coordinates(self):
        return self.get_endpoint_pos()

    def set_intrinsic_coordinates(self, pos):
        self.set_endpoint_pos(pos)

    # def drive(self, decoder):
    #     self.set_intrinsic_coordinates(decoder['q'])
    #     intrinsic_coords = self.get_intrinsic_coordinates()
    #     if not np.any(np.isnan(intrinsic_coords)):
    #         decoder['q'] = self.get_intrinsic_coordinates()
    #     print 'arm pos', self.get_endpoint_pos()

    # def set_endpoint_2D(self, target):
    #     self.system.set_endpoint_2D(target)

    # def set_joints_2D(self, shoulder_angle, elbow_angle): 
    #     self.system.set_joints_2D(shoulder_angle, elbow_angle)

    # def get_hand_location(self, shoulder_anchor):
    #     ''' returns position of ball at end of forearm (hand)'''
    #     return shoulder_anchor + self.system.curr_vecs[0] +self.system.curr_vecs[1]

    # def get_joint_angles_2D(self):
    #     return self.system.curr_angles[0], self.system.curr_angles[1] - np.pi/2





# cursor_bounds = traits.Tuple((-25, 25, 0, 0, -14, 14), "Boundaries for where the cursor can travel on the screen")

chain_kwargs = dict(link_radii=.6, joint_radii=0.6, joint_colors=(181/256., 116/256., 96/256., 1), link_colors=(181/256., 116/256., 96/256., 1))

shoulder_anchor = np.array([2., 0., -15])

chain_15_15_5_5 = RobotArmGen2D(link_lengths=[15, 15, 5, 5], base_loc=shoulder_anchor, **chain_kwargs)
init_joint_pos = np.array([ 0.47515737,  1.1369006 ,  1.57079633,  0.29316668])  ## center pos coordinates: 0.63017,  1.38427,  1.69177,  0.42104 
chain_15_15_5_5.set_intrinsic_coordinates(init_joint_pos)

chain_20_20 = RobotArm2J2D(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)
starting_pos = np.array([5., 0., 5])
chain_20_20.set_endpoint_pos(starting_pos - shoulder_anchor, n_iter=10, n_particles=500)
chain_20_20.set_endpoint_pos(starting_pos, n_iter=10, n_particles=500)

cursor = CursorPlant(endpt_bounds=(-14, 14, 0., 0., -14, 14))
#cursor = CursorPlant(endpt_bounds=(-10, 10, 0., 0., -10, 10))
#cursor = CursorPlant(endpt_bounds=(-9.5, 9.5, 0., 0., -7.5, 11.5))
#cursor = CursorPlant(endpt_bounds=(-11, 11., 0., 0., -11., 11.))
#cursor = CursorPlant(endpt_bounds=(-10, 10., 0., 0., -10., 10.))
#cursor = CursorPlant(endpt_bounds=(-24., 24., 0., 0., -14., 14.))

arm_3d = RobotArm()

plants = dict(RobotArmGen2D=chain_15_15_5_5, 
              RobotArm2J2D=chain_20_20,
              CursorPlant=cursor,
              Arm3D=arm_3d)



'''
Quaternions and generic 3D transformations
'''



import numpy as np

class Quaternion(object):
    def __init__(self, w=1, x=0, y=0, z=0):
        if isinstance(w, (list, tuple, np.ndarray)) and not isinstance(x, np.ndarray):
            #Allows the use of a quaternion as a vector
            self.quat = np.array([0, w[0], w[1], w[2]])
        else:
            self.quat = np.array([w, x, y, z])
    
    def __repr__(self):
        if self.quat.ndim > 1:
            return "<Quaternion set for %d rotations>"%self.quat.shape[1]
        return "%g+%gi+%gj+%gk"%tuple(self.quat)
    
    def norm(self):
        self.quat = self.quat / np.sqrt((self.quat**2).sum())
        return self
    
    def conj(self):
        return Quaternion(self.w, *(-self.vec))

    @property
    def H(self):
        return self.conj()
    
    def __getattr__(self, attr):
        if attr in ["w", "scalar"]:
            return self.quat[0]
        elif attr in ["x", "i"]:
            return self.quat[1]
        elif attr in ["y", "j"]:
            return self.quat[2]
        elif attr in ["z", "k"]:
            return self.quat[3]
        elif attr in ["v", "vec", "vector"]:
            return self.quat[1:]
        else:
            super(Quaternion, self).__getattr__(self, attr)
    
    def __mul__(self, other):
        if isinstance(other, Quaternion):
            w = self.w*other.w   - (self.vec*other.vec).sum(0)
            v = self.w*other.vec + other.w*self.vec + np.cross(self.vec.T, other.vec.T).T
            return Quaternion(w, *v).norm()
        elif isinstance(other, (np.ndarray, list, tuple)):
            if isinstance(other, (list, tuple)):
                other = np.array(other)
            #rotate a vector, will need to be implemented in GLSL eventually
            cross = np.cross(self.vec.T, other) + self.w*other
            return (other + np.cross(2*self.vec.T, cross)).squeeze()
            '''
            conj = self.H
            w = -np.dot(other, conj.vec)
            vec = np.outer(conj.w, other) + np.cross(other, conj.vec.T)
            if self.quat.ndim > 1:
                return self.w*vec.T + np.
            return self.w*vec + np.outer(w, self.vec).squeeze() + np.cross(self.vec, vec)
            '''
        else:
            raise ValueError

    def to_mat(self):
        '''
        Convert to an augmented rotation matrix if the quaternion is of unit norm
        ??? Does this function provide a sensible result for non-unit quaternions?

        Parameters
        ----------
        None

        Returns
        -------
        np.ndarray of shape (4, 4)
            Affine transformation matrix
        '''
        a, b, c, d = self.quat
        return np.array([
            [a**2+b**2-c**2-d**2,       2*b*c-2*a*d,            2*b*d+2*a*c,        0],
            [   2*b*c+2*a*d,       a**2-b**2+c**2-d**2,         2*c*d-2*a*b,        0],
            [   2*b*d-2*a*c,            2*c*d+2*a*b,        a**2-b**2-c**2+d**2,    0],
            [   0,                      0,                          0,              1]])

    @classmethod 
    def from_mat(cls, M):
        qw = np.sqrt(1 + M[0,0] + M[1,1] + M[2,2]) / 2
        qx = (M[2,1] - M[1,2])/(4*qw)
        qy = (M[0,2] - M[2,0])/(4*qw)
        qz = (M[1,0] - M[0,1])/(4*qw)
        return Quaternion(w=qw, x=qx, y=qy, z=qz)

    def rotate_to(self, vec):
        svec = self.vec / np.sqrt((self.vec**2).sum())
        nvec = nvec = vec2 / np.sqrt((vec2**2).sum())
        rad = np.arccos(np.dot(svec, nvec))
        axis = np.cross(svec, nvec)
        self = self.from_axisangle(axis, rad)*self

    @classmethod
    def rotate_vecs(cls, vec1, vec2):
        '''
        Get the quaternion which rotates vec1 onto vec2

        Parameters
        ----------
        vec1: np.ndarray of shape (3,)
            Starting vector
        vec2: np.ndarray of shape (3,)
            Vector which defines the orientation that you want to rotate the first vector to

        Returns
        -------
        Quaternion representing the rotation
        '''
        vec1, vec2 = np.array(vec1), np.array(vec2)
        svec = vec1 / np.sqrt((vec1**2).sum())
        nvec = vec2 / np.sqrt((vec2**2).sum())
        if nvec.ndim > 1:
            if svec.ndim > 1:
                rad = (svec * nvec).sum(1)
            else:
                rad = np.arccos(np.dot(svec, nvec.T))
        else:
            rad = np.arccos(np.dot(svec, nvec))
        axis = np.cross(svec, nvec)
        return cls.from_axisangle(axis, rad)
    
    @classmethod
    def from_axisangle(cls, axis, rad):
        '''
        Convert from the Axis-angle representation of rotations to the quaternion representation

        Parameters
        ----------
        axis: np.ndarray of shape (3,) or ?????
            Rotation axis
        rad: float
            Angle to rotate around the specified axis in radians

        Returns
        -------
        Quaternion representing the rotation
        '''
        #normalize the axis first
        axis = np.array(axis)
        if axis.ndim > 1:
            axis = axis.T / np.sqrt((axis**2).sum(1))
        else:
            if not np.all(axis == 0):
                axis = axis / np.sqrt((axis**2).sum())
        w = np.cos(rad*0.5)
        v = axis * np.sin(rad*0.5)
        return cls(w, *v)

class Transform(object):
    '''
    Homogenous transformations ???
    '''
    def __init__(self, move=(0,0,0), scale=1, rotate=None):
        self.move = np.array(move, dtype=float)
        self.scale = scale
        self.rotate = rotate if rotate is not None else Quaternion()

    def __repr__(self):
        return "Rotate %s, then scale %s, then translate %s"%(self.rotate, self.scale, self.move)
    
    def __mul__(self, other):
        if isinstance(other, Transform):
            #Pre-multiply the other transform, then apply self
            move = self.move + self.rotate*other.move
            scale = self.scale * other.scale
            rot = self.rotate * other.rotate
            return Transform(move, scale, rot)

        elif isinstance(other, Quaternion):
            #Apply the quaternion directly to current rotation
            return Transform(self.move, self.scale, other.rotate * self.rotate)

    def __call__(self, vecs):
        return self.scale * (self.rotate * vecs) + self.move

    def translate(self, x, y, z, reset=False):
        '''
        Set the translation point of the transformation

        Parameters
        ----------
        x, y, z: float
            Coordinates representing how much to move
        reset: bool, optional, default=False 
            If true, the new coordinates replace the old ones. If false, they are added on
        '''
        if reset:
            self.move[:] = x,y,z
        else:
            self.move += x,y,z
        return self

    def rotate_x(self, rad, reset=False):
        rotate = Quaternion.from_axisangle((1,0,0), rad)
        if reset:
            self.rotate = rotate
        else:
            self.rotate = (rotate * self.rotate).norm()
        return self

    def rotate_y(self, rad, reset=False):
        rotate = Quaternion.from_axisangle((0,1,0), rad)
        if reset:
            self.rotate = rotate
        else:
            self.rotate = (rotate * self.rotate).norm()
        return self

    def rotate_z(self, rad, reset=False):
        rotate = Quaternion.from_axisangle((0,0,1), rad)
        if reset:
            self.rotate = rotate
        else:
            self.rotate = (rotate * self.rotate).norm()
        return self
    
    def to_mat(self, reverse=False):
        scale = np.eye(4)
        scale[(0,1,2), (0,1,2)] = self.scale
        move = np.eye(4)
        move[:3, -1] = self.move
        
        if reverse:
            return np.dot(self.rotate.to_mat(), np.dot(scale, move))
        else:
            return np.dot(move, np.dot(scale, self.rotate.to_mat()))

def test():
    world = Transform().rotate_x(np.radians(-90))
    eye = Transform().translate(0,35,0)
    obj = Transform().translate(0,10,5)
    assert np.allclose((world*eye*obj)((0,0,1)), [0,6,-45])
    obj.rotate_y(np.radians(-90))
    assert np.allclose((world*eye*obj)((0,0,1)), [-1, 5, -45])
    obj.rotate_z(np.radians(-90))
    assert np.allclose((world*eye*obj)((0,0,1)), [0,5,-46])
    assert np.allclose(np.dot((world*eye*obj).to_mat(), [0,0,1,1]), [0,5,-46, 1])



'''
Graphical display classes. Experimental tasks involving graphical displays
inherit from these classes.
'''
import os

import numpy as np
from OpenGL.GL import *

from ..experiment import LogExperiment
from ..experiment import traits

from .render import stereo, render, ssao, shadow_map
from .models import Group
from .xfm import Quaternion, Transform
from .primitives import Sphere, Cube, Chain
from .environment import Box, Grid
from .primitives import Cylinder, Sphere, Cone
import socket

try:
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
    import pygame
except ImportError:
    import warnings
    warnings.warn('riglib/stereo_opengl/window.py: not importing name pygame')

from config.rig_defaults import window as defaults

# for WindowDispl2D only
from .primitives import Shape2D

class Window(LogExperiment):
    '''
    Generic stereo window
    '''
    status = dict(draw=dict(stop=None))
    state = "draw"
    stop = False

    # XPS computer
    # window_size = (1920*2, 1080)
    window_size = traits.Tuple(defaults['window_size'], desc='Window size, in pixels')
    background = traits.Tuple((0.,0.,0.,1.), desc="Background color (R,G,B,A)")
    fullscreen = traits.Bool(True, desc="Fullscreen window")

    #Screen parameters, all in centimeters -- adjust for monkey
    screen_dist = traits.Float(defaults['screen_dist'], desc="Screen to eye distance (cm)")
    screen_half_height = traits.Float(defaults['screen_half_height'], desc="Screen half height (cm)")
    iod = traits.Float(2.5, desc="Intraocular distance (cm)")     # intraocular distance
    stereo_mode = traits.OptionsList(['hmd', 'mirror', 'projection', 'anaglyph'], desc="Stereo mode", 
                                     bmi3d_input_options=['hmd', 'mirror', 'projection', 'anaglyph'])

    show_environment = traits.Bool(False, desc="Show wireframe box around environment")

    hidden_traits = ['stereo_mode', 'screen_dist', 'screen_half_height', 'iod', 'show_environment', 'background']

    def __init__(self, *args, **kwargs):
        self.display_start_pos = kwargs.pop('display_start_pos', "0,0")
        super(Window, self).__init__(*args, **kwargs)

        self.models = []
        self.world = None
        self.event = None

        # os.popen('sudo vbetool dpms on')
        self.fov = np.degrees(np.arctan(self.screen_half_height/self.screen_dist))*2
        self.screen_cm = [2 * self.screen_half_height * self.window_size[0]/self.window_size[1], 2 * self.screen_half_height]

    def set_os_params(self):
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"

    def screen_init(self):
        self.set_os_params()
        pygame.init()

        pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 24)
        pygame.display.gl_set_attribute(pygame.GL_FRAMEBUFFER_SRGB_CAPABLE, True)
        pygame.display.gl_set_attribute(pygame.GL_CONTEXT_MAJOR_VERSION, 3)
        pygame.display.gl_set_attribute(pygame.GL_CONTEXT_MINOR_VERSION, 2)
        pygame.display.gl_set_attribute(pygame.GL_CONTEXT_PROFILE_MASK, pygame.GL_CONTEXT_PROFILE_CORE)
        flags = pygame.DOUBLEBUF | pygame.HWSURFACE | pygame.OPENGL | pygame.NOFRAME
        if self.fullscreen:
            flags = flags | pygame.FULLSCREEN
        try:
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS,1)
            self.screen = pygame.display.set_mode(self.window_size, flags)
        except:
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS,0)
            self.screen = pygame.display.set_mode(self.window_size, flags)

        glDisable(GL_FRAMEBUFFER_SRGB) # disable gamma correction
        glEnable(GL_BLEND)
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(*self.background)
        glClearDepth(1.0)
        glDepthMask(GL_TRUE)
        glEnable(GL_CULL_FACE) # temporary solution to alpha blending issue with spheres. just draw the front half of the sphere
        glCullFace(GL_BACK)


        #this effectively determines the modelview matrix
        self.world = Group(self.models)
        self.world.init()

        #up vector is always (0,0,1), why would I ever need to roll the camera?!
        self.set_eye((0, 0, 0), (0,0))
        self.modelview = np.eye(4)
        self.modelview[:3,-1] = [0,0,-self.screen_dist]
        self.renderer = self._get_renderer()

        pygame.mouse.set_visible(False)

        if self.show_environment:
            self.add_model(Box())
    
    def _get_renderer(self):
        near = 1
        far = 1024
        if self.stereo_mode == 'mirror':
            self.modelview = np.dot(self.modelview, np.diag([-1,1,1,1]))
            glFrontFace(GL_CW);  # Switch to clockwise winding for mirrored objects
            return stereo.MirrorDisplay(self.window_size, self.fov, near, far, self.screen_dist, self.iod)
        if self.stereo_mode == 'hmd':
            return stereo.MirrorDisplay(self.window_size, self.fov, near, far, self.screen_dist, self.iod)
        elif self.stereo_mode == 'projection':
            return shadow_map.ShadowMapper(self.window_size, self.fov, near, far)
        elif self.stereo_mode == 'anaglyph':
            return stereo.Anaglyph(self.window_size, self.fov, near, far, self.screen_dist, self.iod)
        else:
            raise ValueError("Unknown stereo mode: %s" % self.stereo_mode)

    def set_eye(self, pos, vec, reset=True):
        '''Set the eye's position and direction. Camera starts at (0,0,0), pointing towards positive y'''
        self.world.translate(pos[0], pos[2], pos[1], reset=True).rotate_x(-90)
        self.world.rotate_y(vec[0]).rotate_x(vec[1])

    def add_model(self, model):
        if self.world is None:
            #world doesn't exist yet, add the model to cache
            self.models.append(model)
        else:
            #We're already running, initialize the model and add it to the world
            model.init()
            self.world.add(model)

    def remove_model(self, model):
        '''remove the second model from the last'''
        self.world.remove(model)

    def show_object(self, obj, show=False):
        '''
        Show or hide an object. This function is an abstraction so that tasks don't need to know about attach/detach
        '''
        if show:
            obj.attach()
        else:
            obj.detach()

    def draw_world(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.renderer.draw(self.world, modelview=self.modelview)
        pygame.display.flip()
        self.renderer.draw_done()

    def _get_event(self):
        for e in pygame.event.get(pygame.KEYDOWN):
            return (e.key, e.type)

    def _start_None(self):
        pygame.display.quit()

    def _test_stop(self, ts):
        '''
        Stop the task if the escape key is pressed, or if the super _test_stop instructs a stop
        '''
        super_stop = super(Window, self)._test_stop(ts)
        from pygame import K_ESCAPE
        key_stop = self.event is not None and self.event[0] == K_ESCAPE
        if key_stop:
            print("Window closed. Stopping task")
        return super_stop or key_stop
    
    def update_report_stats(self):
        self.reportstats['FPS'] = round(self.clock.get_fps(),2)
        return super().update_report_stats()
    
    def requeue(self):
        self.renderer._queue_render(self.world)

    def _cycle(self):
        super(Window, self)._cycle() # is this order intentional? why not cycle first then draw the screen?
        self.event = self._get_event()
        self.requeue()
        self.draw_world()

    
class WindowSSAO(Window):
    '''Screen Space Ambient Occlusion (SSAO) window. Replaces the renderer with one using SSAO'''

    def _get_renderer(self):
        return ssao.SSAO(self.window_size, self.fov, 0.05, 1024)


class WindowWithExperimenterDisplay(Window):
    hostname = socket.gethostname()
    # This class has a hard-coded window size
    if hostname == 'lynx':
        _stereo_window_flip = True
        _stereo_main_flip_z = True
        window_size = (1280+480, 1024)
        window_size1 = (1280, 1024)
        window_size2 = (480, 270)
    else:
        _stereo_window_flip = False
        _stereo_main_flip_z = False
        window_size = (1920 + 480, 1080)
        window_size1 = (1920, 1080)
        window_size2 = (480, 270)

    def __init__(self, *args, **kwargs):
        super(WindowWithExperimenterDisplay, self).__init__(*args, **kwargs)
        # This class has a hard-coded window size
        # self.window_size = (1920 + 480, 1080)

    def set_os_params(self):
        # NOTE: in Ubuntu Unity, setting the SDL_VIDEO_WINDOW_POS seems to be largely ignored.
        # You can set which screen the window appears on if you have a dual display, but you cannot set the exact position
        # Instead, you have to hard-code a render start location in the compiz-config settings manager
        # http://askubuntu.com/questions/452995/how-to-adjust-window-placement-in-unity-ubuntu-14-04-based-on-overlapping-top-b
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment_with_mini"

    def _get_renderer(self):
        near = 1
        far = 1024
        return stereo.DualMultisizeDisplay(self.window_size1, self.window_size2, self.fov, near, far, self.screen_dist, self.iod, flip=self._stereo_window_flip,
            flip_main_z = self._stereo_main_flip_z)


TRANSPARENT = (106,0,42)

class WindowDispl2D():
    '''For testing only -- draws world on a 2D screen. May cause mild confusion -- transforms 
    incoming 3D coordinates (x,y,z) into 2D coordinates (x,y) by mapping z onto y'''

    def __init__(self, *args, **kwargs):
        self.is_pygame_display = True
        super().__init__(*args, **kwargs)

    def screen_init(self):
        os.environ['SDL_VIDEO_WINDOW_POS'] = self.display_start_pos
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"
        pygame.init()
        self.clock = pygame.time.Clock()

        flags = pygame.NOFRAME | pygame.DOUBLEBUF | pygame.HWSURFACE

        if self.fullscreen:
            flags = flags | pygame.FULLSCREEN

        self.display_border = 10

        self.size = np.array(self.window_size, dtype=np.float64)
        self.screen = pygame.display.set_mode(self.window_size, flags)
        self.screen_background = pygame.Surface(self.screen.get_size()).convert()
        self.screen_background.fill(self.background)

        x_max, y_max = (self.screen_cm[0]/2, self.screen_cm[1]/2)
        x_min, y_min = (-self.screen_cm[0]/2, -self.screen_cm[1]/2)
        self.normalize = np.array(np.diag([1./(x_max-x_min), 1./(y_max-y_min), 1]))
        self.center_xform = np.array([[1., 0, -x_min], 
                                      [0, 1., -y_min],
                                      [0, 0, 1]])
        self.norm_to_screen = np.array(np.diag(np.hstack([self.size, 1])))

        # the y-coordinate in pixel space has to be swapped graphics convention reasons
        self.flip_y_coord = np.array([[1, 0, 0],
                                      [0, -1, self.size[1]],
                                      [0, 0, 1]])
        self.pos_space_to_pixel_space = np.dot(self.flip_y_coord, np.dot(self.norm_to_screen, np.dot(self.normalize, self.center_xform)))

        self.world = Group(self.models)
        # Dont 'init' self.world in this Window. Just allocates a bunch of OpenGL stuff which is not necessary (and may not work in some cases)
        # self.world.init()

        #initialize surfaces for translucent markers
        self.surf={}
        self.surf['0'] = pygame.Surface(self.screen.get_size())
        self.surf['0'].fill(TRANSPARENT)
        self.surf['0'].set_colorkey(TRANSPARENT)

        self.surf['1'] = pygame.Surface(self.screen.get_size())
        self.surf['1'].fill(TRANSPARENT)
        self.surf['1'].set_colorkey(TRANSPARENT)

         #values of alpha: higher = less translucent
        self.surf['0'].set_alpha(170) #Cursor
        self.surf['1'].set_alpha(130) #Targets

        self.surf_background = pygame.Surface(self.surf['0'].get_size()).convert()
        self.surf_background.fill(TRANSPARENT)

        pygame.mouse.set_visible(False)

        self.i = 0

    def pos2pix(self, kfpos):
        # re-specify the point in homogenous coordinates
        pt = np.hstack([kfpos, 1]).reshape(-1, 1)

        # perform the homogenous transformation
        pix_coords = np.dot(self.pos_space_to_pixel_space, pt)

        pix_pos = np.array(pix_coords[:2,0], dtype=int)
        return pix_pos

    def get_surf(self):
        return self.surf[str(np.min([self.i,1]))]

    def draw_model(self, model):
        '''
        Draw a single Model on the current surface, or recurse if the model is a composite model (i.e., a Group)
        '''
        color = tuple([int(255*x) for x in model.color[0:3]])
        if isinstance(model, Sphere):
            pos = model._xfm.move[[0,2]]
            pix_pos = self.pos2pix(pos)

            rad = model.radius
            pix_radius = self.pos2pix(np.array([model.radius, 0]))[0] - self.pos2pix([0,0])[0]

            #Draws cursor and targets on transparent surfaces
            pygame.draw.circle(self.get_surf(), color, pix_pos, pix_radius)

        elif isinstance(model, Shape2D):
            # model.draw() returns True if the object was drawn
            #   (which happens if the object's .visible attr is True)
            if model.draw(self.get_surf(), self.pos2pix):
                pass
        elif isinstance(model, Cube):
            pos = model.xfm.move[[0,2]]
            side_len = model.side_len

            left = pos[0] - side_len/2
            right = pos[0] + side_len/2
            top = pos[1] + side_len/2
            bottom = pos[1] - side_len/2

            top_left = np.array([left, top])
            bottom_right = np.array([right, bottom])
            top_left_pix_pos = self.pos2pix(top_left)
            bottom_right_pix_pos = self.pos2pix(bottom_right)

            rect = pygame.Rect(top_left_pix_pos, bottom_right_pix_pos - top_left_pix_pos)
            color = tuple([int(255*x) for x in model.color[0:3]])

            pygame.draw.rect(self.get_surf(), color, rect)

        elif isinstance(model, (Cylinder, Cone)):
            vec_st = np.array([0., 0, 0, 1])
            vec_end = np.array([0., 0, model.height, 1])

            cyl_xform = model._xfm.to_mat()
            cyl_start = np.dot(cyl_xform, vec_st)
            cyl_end = np.dot(cyl_xform, vec_end)
            pix_radius = self.pos2pix(np.array([model.radius, 0]))[0] - self.pos2pix([0,0])[0]

            pygame.draw.line(self.get_surf(), color, self.pos2pix(cyl_start[[0,2]]), self.pos2pix(cyl_end[[0,2]]), pix_radius)

            # print cyl_start, cyl_end

        elif isinstance(model, Group):
            for mdl in model:
                self.draw_model(mdl)

    def _draw_other(self):
        pass

    def draw_world(self):
        #Refreshes the screen with original background
        self.screen.blit(self.screen_background, (0, 0))
        self.surf['0'].blit(self.surf_background,(0,0))
        self.surf['1'].blit(self.surf_background,(0,0))

        # surface index
        self.i = 0

        for model in self.world.models:
            self.draw_model(model)
            self.i += 1

        #Renders the new surfaces
        self.screen.blit(self.surf['0'], (0,0))
        self.screen.blit(self.surf['1'], (0,0))

        self._draw_other()
        pygame.display.update()

    def requeue(self):
        '''
        Simulation 'requeue' does nothing because the simulation is lazy and
        inefficient and chooses to redraw the entire screen every loop
        '''
        pass


class Window2D():
    '''
    Render the world in 2D without lighting
    '''
    def _get_renderer(self):
        glCullFace(GL_BACK)
        return render.Renderer2D(self.screen_cm)
    
class FakeWindow(Window):
    '''
    A dummy class to secretly avoid rendering graphics without
    the graphics-based tasks knowing about it. Used e.g. for simulation
    purposes where the graphics only slow down the simulation.
    '''
    background = (1,1,1,1)
    def __init__(self, *args, **kwargs):
        self.models = []
        self.world = None
        self.event = None
        super(FakeWindow, self).__init__(*args, **kwargs)

    def screen_init(self):
        self.world = Group(self.models)
        # self.world.init()

    def draw_world(self):
        pass

    def requeue(self):
        pass

    def _start_reward(self, *args, **kwargs):
        n_rewards = self.calc_state_occurrences('reward')
        if n_rewards % 10 == 0:
            print(n_rewards)
        super(FakeWindow, self)._start_reward(*args, **kwargs)

    def show_object(self, *args, **kwargs):
        pass

    def _get_event(self):
        pass


class FPScontrol(Window):
    '''A mixin that adds a WASD + Mouse controller to the window.
    Use WASD to move in XY plane, q to go down, e to go up'''

    def screen_init(self):
        super(FPScontrol, self).screen_init()
        pygame.mouse.set_visible(False)
        self.eyepos = [0,-self.screen_dist, 0]
        self.eyevec = [0,0]
        self.wasd = [False, False, False, False, False, False]

    def _get_event(self):
        retme = None
        for e in pygame.event.get([pygame.MOUSEMOTION, pygame.KEYDOWN, pygame.KEYUP, pygame.QUIT]):
            moved = any(self.wasd)
            if e.type == pygame.MOUSEMOTION:
                self.world.xfm.rotate *= Quaternion.from_axisangle((1,0,0), np.radians(e.rel[1]*.1))
                self.world.xfm.rotate *= Quaternion.from_axisangle((0,0,1), np.radians(e.rel[0]*.1))
                self.world._recache_xfm()
            elif e.type == pygame.KEYDOWN:
                kn = pygame.key.name(e.key)
                if kn in ["escape", "q"]:
                    self.stop = True
                retme = (e.key, e.type)
            elif e.type == pygame.QUIT:
                self.stop = True

            if moved:
                self.set_eye(self.eyepos, self.eyevec, reset=True)
        return retme



# Generated by Django 2.2.1 on 2020-07-27 14:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0004_keyvaluestore'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='sw_version',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



'''
An extension of Tornado's web socket which enables the task to
print data to the web interface while the task is running.
'''

import os
import sys
import cgi
import json
import struct
import multiprocessing as mp
import tornado.ioloop
import tornado.web
from tornado import websocket
import io, traceback
import copy

sockets = []

class ClientSocket(websocket.WebSocketHandler):
    def open(self):
        sockets.append(self)
        print("WebSocket opened")

    def on_close(self):
        print("WebSocket closed")
        sockets.remove(self)

    def check_origin(self, origin):
        '''
        Returns a boolean indicating whether the requesting URL is one that the
        handler will respond to. For this websocket, everyone with access gets a response since
        we're running the server locally (or over ssh tunnels) and not over the regular internet.

        Parameters
        ----------
        origin : string?
            The URL from which the request originated

        Returns
        -------
        boolean
            Returns True if the request originates from a valid URL

        See websocket.WebSocketHandler.check_origin for additional documentation
        '''
        return True


class Server(mp.Process):
    '''
    Spawn a process to deal with the websocket asynchronously, without halting other webserver operations.
    '''
    def __init__(self, notify=None):
        super(self.__class__, self).__init__()
        self._pipe, self.pipe = os.pipe()
        self._outp, self.outp = os.pipe()
        self.outqueue = ""
        self.notify = notify
        self.start()

    def run(self):
        '''
        Main function to run in the process. See mp.Process.run() for additional documentation.
        '''
        print("Running websocket service")
        application = tornado.web.Application([
            (r"/connect", ClientSocket),
        ])

        import asyncio
        asyncio.set_event_loop(asyncio.new_event_loop())

        application.listen(8001)
        self.ioloop = tornado.ioloop.IOLoop.current()
        self.ioloop.add_handler(self._pipe, self._send, self.ioloop.READ)
        self.ioloop.add_handler(self._outp, self._stdout, self.ioloop.READ)
        self.ioloop.start()

    def send(self, msg):
        # notify the task tracker that data is being sent to the web interface
        if self.notify is not None:
            self.notify(msg)

        # force the message to string, if necessary (e.g., NotifyFeat.set_state sends a dict)
        if not isinstance(msg, str):
            msg = json.dumps(msg)

        # Write to 'self.pipe'. The write apparently triggers the function self._stdout to run
        os.write(self.pipe, struct.pack('I', len(msg)) + bytes(msg, 'utf8'))

    def _stdout(self, fd, event):
        '''
        Handler for self._pipe; Read the data from the input pipe and propagate the data to all the listening sockets
        '''
        nbytes, = struct.unpack('I', os.read(fd, 4))
        msg = os.read(fd, nbytes)

        # write the message to all the sockets--could be multiple websockets open if multiple people are observing the same experiment
        for sock in sockets:
            # each 'sock' is assumed to be an instance of the ClientSocket above
            sock.write_message(msg)

    def stop(self):
        print("Stopping websocket service")
        self.send(dict(status="stopped", State="stopped"))
        self.join()

    ##### Currently unused functions below this line #####
    def write(self, data):
        '''Used for stdout hooking'''
        self.outqueue += data
        self.flush()

    def flush(self):
        msg = json.dumps(dict(status="stdout", msg=cgi.html.escape(self.outqueue)))
        os.write(self.outp, struct.pack('I', len(msg)) + bytes(msg, 'utf8'))
        self.outqueue = ""

    def _send(self, fd, event):
        nbytes, = struct.unpack('I', os.read(fd, 4))
        msg = os.read(fd, nbytes)
        if msg == "stop":
            self.ioloop.stop()
        else:
            for sock in sockets:
                sock.write_message(msg)


class NotifyFeat(object):
    '''
    Send task report and state data to display on the web inteface
    '''
    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)
        self.websock = kwargs.pop('websock')
        self.tracker_status = kwargs.pop('tracker_status')
        self.prev_stats = None

    def _cycle(self):
        super()._cycle()
        self.reportstats['status'] = str(self.tracker_status)
        if self.reportstats != self.prev_stats:
            self.websock.send(self.reportstats)
            self.prev_stats = copy.deepcopy(self.reportstats)

    def run(self):
        try:
            super().run()
        except:
            err = io.StringIO()
            traceback.print_exc(None, err)
            err.seek(0)
            self.websock.send(dict(status="error", msg=err.read()))
        finally:
            if self.terminated_in_error:
                self.websock.send(dict(status="error", msg=self.termination_err.read()))
            else:
                self.reportstats['status'] = str(self.tracker_status)
                self.websock.send(self.reportstats)

    def print_to_terminal(self, *args):
        sys.stdout = sys.__stdout__
        super().print_to_terminal(*args)
        sys.stdout = self.websock


class WinNotifyFeat(object):
    '''
    Stop the task gracefully on windows without websockets
    '''
    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)
        self.tracker_end_of_pipe = kwargs.pop('tracker_end_of_pipe')
        self.tracker_status = kwargs.pop('tracker_status')

    def run(self):
        try:
            super().run()
        except:
            err = io.StringIO()
            traceback.print_exc(None, err)
            err.seek(0)
            print(err.read())
        finally:
            self.tracker_end_of_pipe.send(None)



'''
This module contains functions which convert parameter sets stored
as JSON blobs into python dictionaries or vice versa.
'''
import builtins
import ast
import json
import numpy as np
import os

log_path = os.path.join(os.path.dirname(__file__), '../../log')


def param_objhook(obj):
    '''
    A custom JSON "decoder" which can recognize certain types of serialized python objects
    (django models, function calls, object constructors) and re-create the objects

    Parameters
    ----------
    obj : dict
        The deserialized JSON data in python dictionary form (after calling json.loads)

    Returns
    -------
    object
        If possible, a python object based on the JSON data is created. If not, the original dictionary
        is simply returned.

    '''
    from . import models
    if '__django_model__' in obj:
        model = getattr(models, obj['__django_model__'])
        return model(pk = obj['pk'])
    elif '__builtin__' in obj:
        func = getattr(builtins, obj['__builtin__'])
        return func(*obj['args'])
    elif '__class__' in obj:
        # look up the module
        mod = builtins.__import__(obj['__module__'], fromlist=[obj['__class__']])

        # get the class with the 'getattr' and then run the class constructor on the class data
        return getattr(mod, obj['__class__'])(obj['__dict__'])
    else: # the type of object is unknown, just return the original dictionary
        return obj

def norm_trait(trait, value):
    '''
    Take user input and convert to the type of the trait.
    For example, a user might select a decoder's name/id but the ID needs to be mapped
    to an object for type checking when the experiment is constructed)

    Parameters
    ----------
    trait : trait object
        trait object declared for the runtime-configurable field.
    value : object
        Value of trait set from user input, to be type-checked

    Returns
    -------
    typecast value of trait
    '''
    ttype = trait.trait_type.__class__.__name__
    if ttype == 'Instance':
        # if the trait is an 'Instance' type and the value is a number, then the number gets interpreted as the primary key to a model in the database
        if isinstance(value, int):
            cname = namelist.instance_to_model[trait.trait_type.klass]
            record = cname.objects.get(pk=value)
            value = record.get()
        # Otherwise, let's hope it's already an instance
    elif ttype == 'InstanceFromDB':
        if isinstance(value, int):
            # look up the model name in the trait
            mdl_name = trait.bmi3d_db_model
            # get the database Model class from 'db.tracker.models'
            with open(os.path.join(log_path, "json_param_log"), "w") as f:
                f.write(str(trait) + "\n")
                f.write(str(mdl_name) + "\n")

            from . import models
            Model = getattr(models, mdl_name)
            record = Model.objects.get(pk=value)
            value = record.get()
        # Otherwise, let's hope it's already an instance
    elif ttype == 'DataFile':
        # Similar to Instance traits, except we always know to use models.DataFile as the database table to look up the primary key
        from . import models
        if isinstance(value, int):
            record = models.DataFile.objects.get(pk=value)
            value = record.get()
    elif ttype == 'Bool':
        if value == 'on':
            value = True
        elif value == 'off':
            value = False
        elif value == 'true':
            value = True
        elif value == 'false':
            value = False
    elif ttype == 'Tuple':
        # Explicit cast to tuple for backwards compatibility reasons (should not be necessary for newer versions of the code/traits lib?)
        value = tuple(value)
    else:
        if isinstance(value, str):
            value = _parse_str(value)

    #use Cast to validate the value
    try:
        if trait.cast is not None:
            return trait.cast
        else:
            # this interface changed in later versions of the trait library
            return trait.validate(trait, 'casting', value)
    except:
        f = open(os.path.join(log_path, "trait_log"), 'w')
        f.write('Error with type for trait %s, %s, value %s' % (str(trait), str(ttype), str(value)))
        f.close()
        import traceback
        traceback.print_exc()
        raise ValueError("Invalid input for parameter %s: %s" % (str(trait.name), str(value)))


def _parse_str(value):
    try:
        return json.loads(value, object_hook=param_objhook)
    except:
        try:
            return ast.literal_eval(value)
        except:
            return value;

class Parameters(object):
    def __init__(self, rawtext):
        if rawtext == '':
            rawtext = '{}'
        self.params = json.loads(rawtext, object_hook=param_objhook)

    @classmethod
    def from_dict(cls, params):
        c = cls('null')
        c.params = params
        return c

    @classmethod
    def from_html(cls, params):
        processed = dict()
        for name, value in list(params.items()):
            if isinstance(value, str):
                processed[name] = _parse_str(value)
            elif isinstance(value, list):
                processed[name] = list(map(_parse_str, value))
            else:
                processed[name] = value

        return cls.from_dict(processed)

    def to_json(self):
        from django.db import models
        def encode(obj):
            if isinstance(obj, models.Model):
                # If an object is a Django model instance, serialize it using just the model name and the primary key
                return dict(
                    __django_model__=obj.__class__.__name__,
                    pk=obj.pk)
            elif isinstance(obj, tuple):
                # for some reason, the re-constructor needs to specified as a tuple?
                return dict(__builtin__="tuple", args=[obj])
            elif isinstance(obj, np.ndarray):
                # serialize numpy arrays as lists
                return obj.tolist()
            elif isinstance(obj, dict):
                # if the object is a dictionary, just run the encoder on each of the 'values' of the dictionary
                return dict((k, encode(v)) for k, v in list(obj.items()))
            elif isinstance(obj, object) and hasattr(obj, '__dict__'):
                # if the object is a new-style class (inherits from 'object'), save the module, class name and object data
                # (python object data (attributes) are stored in the parameter __dict__)
                data = dict(
                    __module__=obj.__class__.__module__,
                    __class__=obj.__class__.__name__,
                    __dict__=obj.__dict__)

                if hasattr(obj, '__getstate__'):
                    data['__dict__'] = obj.__getstate__()
                return data
            else:
                return obj

        return json.dumps(encode(self.params))

    def trait_norm(self, traits):
        '''
        Apply typecasting to parameters which correspond to experiment traits

        Parameters
        ----------
        traits : dict
            keys are the names of each trait for the Experiment class, values are the trait objects

        Returns
        -------
        None
        '''
        params = self.params
        self.params = dict()
        for name, value in list(params.items()):
            if name in traits:
                traits[name].name = name
                self.params[name] = norm_trait(traits[name], value)
            else:
                self.params[name] = value

    def get_data(self):
        """ Return data with values converted to numbers if possible """
        params_parsed = dict()
        for key in self.params:
            # The internet suggests this might be the best way to check
            # if a string is a number...
            try:
                params_parsed[key] = float(self.params[key])
            except:
                params_parsed[key] = self.params[key]
        return params_parsed

    def __contains__(self, attr):
        return attr in self.params

    def __getitem__(self, attr):
        return self.params[attr]

    def __setitem__(self, attr, val):
        self.params[attr] = val



'''
HDF-saving features
'''
import tempfile
import os
from riglib import sink
from riglib.experiment.experiment import control_decorator
import hdfwriter # from riglib.hdfwriter import hdfwriter # <-- now a package on PyPI

log_path = os.path.join(os.path.dirname(__file__), '../log/hdf_sink.log')

class SaveHDF():
    '''
    Saves data from registered sources into tables in an HDF file
    '''

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' starts an HDFWriter sink.
        '''
        self.h5file = tempfile.NamedTemporaryFile(suffix=".h5", dir='/var/tmp', delete=False)
        self.h5file.flush()
        self.h5file.close()

        sink_manager = sink.SinkManager.get_instance()
        self.hdf = sink_manager.start(hdfwriter.HDFWriter, filename=self.h5file.name, log_filename=log_path)

        self.h5file_name = self.h5file.name

        super(SaveHDF, self).init()    

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method stops the HDF sink after the FSM has finished running        
        '''
        try:
            super(SaveHDF, self).run()
        finally:
            self.hdf.stop()
    
    def join(self):
        '''
        Re-join any spawned process for cleanup
        '''
        self.hdf.join()
        super(SaveHDF, self).join()

    def set_state(self, condition, **kwargs):
        '''
        Save task state transitions to HDF

        Parameters
        ----------
        condition: string
            Name of new state to transition into. The state name must be a key in the 'status' dictionary attribute of the task

        Returns
        -------
        None
        '''
        self.hdf.sendMsg(condition)
        super(SaveHDF, self).set_state(condition, **kwargs)

    @control_decorator
    def record_annotation(self, msg=""):
        """ Record a user-input annotation """
        self.hdf.sendMsg("annotation: " + msg)
        print("Saved annotation to HDF: " + msg)

    def get_h5_filename(self):
        return self.h5file.name        

    def cleanup(self, database, saveid, **kwargs):
        '''
        See LogExperiment.cleanup for documentation
        '''
        super(SaveHDF, self).cleanup(database, saveid, **kwargs)
        print("Beginning HDF file cleanup")
        print("\tHDF data currently saved to temp file: %s" % self.h5file.name)
        try:
            print("\tRunning self.cleanup_hdf()")
            self.cleanup_hdf()
        except:
            print("\n\n\n\n\nError cleaning up HDF file!")
            import traceback
            traceback.print_exc()

        # this 'if' is needed because the remote procedure call to save_data doesn't like kwargs
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        if dbname == 'default':
            database.save_data(self.h5file.name, "hdf", saveid)
        else:
            database.save_data(self.h5file.name, "hdf", saveid, dbname=dbname)



import numpy as np
from riglib.experiment import traits
from riglib.bmi import clda

import aopy
import glob
import os



class CLDA_KFRML_IntendedVelocity(traits.HasTraits):
    clda_batch_time = traits.Float(1, desc="How frequently to update weights [s]")
    clda_update_half_life = traits.Float(60, desc="Half-life for exponential decay [s] to combine with previous weights.") #[s]
    # clda_update_batch_time = traits.Float(60, desc="How frequently to update weights [s]")
    # clda_learner_batch_time = traits.Float(60, desc="How much data to update the learner with [s]") # Samples to update intended kinematics with
    def create_learner(self):
        '''
        The "learner" uses knowledge of the task goals to determine the "intended"
        action of the BMI subject and pairs this intention estimation with actual observations.
        '''
        self.learn_flag = False
        fmatrix = np.array(self.decoder.filt.B.T/np.max(self.decoder.filt.B))
        self.decoder.filt.F_dict = {
            'target': fmatrix,
            'hold': np.zeros(fmatrix.shape),
            'timeout_penalty': np.zeros(fmatrix.shape),
            'wait': np.zeros(fmatrix.shape),
            'delay': np.zeros(fmatrix.shape),
            'targ_transition': np.zeros(fmatrix.shape),
            'hold_penalty': np.zeros(fmatrix.shape),
            'delay_penalty': np.zeros(fmatrix.shape),
            'reward': np.zeros(fmatrix.shape),
        }

        learner_batch_size = int(self.clda_batch_time/self.decoder.binlen)
        self.learner = clda.OFCLearnerRotateIntendedVelocity(learner_batch_size, self.decoder.filt.A, self.decoder.filt.B, self.decoder.filt.F_dict)

    def create_updater(self):
        '''
        The "updater" uses the output batches of data from the learner and an update rule to
        alter the decoder parameters to better match the intention estimates.
        '''
        self.updater = clda.KFRML(self.clda_batch_time, self.clda_update_half_life)
        self.updater.init(self.decoder)


"""Source template for a serial port streaming sensor, e.g., from an Arduino"""
import serial
import numpy as np
import traceback
import time

from riglib.source import DataSourceSystem

class SerialPortSource(DataSourceSystem):
    default_response = []
    dtype = np.dtype([("sensor_data", float)])
    update_freq = 100

    def __init__(self, port, baudrate):
        self.port = serial.Serial(port, baudrate)
        self.port.flushInput()
        self.port.reset_output_buffer()

    def start(self):
        self.port.flushInput()
        self.port.reset_output_buffer()

    def stop(self):
        self.port.flushInput()
        self.port.reset_output_buffer()
        self.port.close()

    def get(self):
        '''
        Retrieve the current data available from the source. 
        '''
        data = self.default_response
        try:
            data_raw = self.port.readline().rstrip().decode('utf-8')
            data = np.zeros((1,), dtype=self.dtype)
            data_vals = []
            for k, x in enumerate(data_raw.split(', ')):
                try:
                    x = float(x)
                except:
                    pass
                data_vals.append(x)
            data_vals = tuple(data_vals)

            if len(self.dtype) == 1 and len(data_vals) > 1:
                data[self.dtype.names[0]] = data_vals
            elif len(self.dtype) == len(data_vals) + 1 and self.dtype.names[0] == 'rxts':
                data['rxts'] = time.time()
                for name, val in zip(self.dtype.names[1:], data_vals):
                    data[name] = val                
            elif len(self.dtype) == len(data_vals):
                for name, val in zip(self.dtype.names, data_vals):
                    data[name] = val
        except:
            traceback.print_exc()
            
        return data

class SerialPortFeature(object):
    pass